"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import List, Tuple

SQRT5 = math.sqrt(5)
INV_SQRT5 = math.sqrt(0.2)

# Dodecahedron face centers (origins) can be defined exactly using trigonometry
# The north and south poles are just at z=1 and z=-1
# Then there are two rings at z = ±INV_SQRT5, with radius 2 * INV_SQRT5

# Exact values for defining a regular pentagon (with radius 1). It is correct to use a radius
# of 1 as we want to obtain the axes of rotations, so the vectors need to be normalized.
# cos0 = 0;
# cos36 = (SQRT5 + 1) / 4;
# cos72 = (SQRT5 - 1) / 4;
# sin0 = 0;
# sin36 = Math.sqrt(10 - 2 * SQRT5) / 4;
# sin72 = Math.sqrt(10 + 2 * SQRT5) / 4;
#
# To compute the quaternion use the equation:
# q = [...sin(alpha) * axis, cos(alpha)]
# where alpha is the half-angle of rotation from the pole to the face center.

# Sin/cosine of half angle (alpha) of rotation from pole to first ring
# For the second ring sin -> cos and cos -> -sin by (pi / 2 - x) identities
sin_alpha = math.sqrt((1 - INV_SQRT5) / 2)
cos_alpha = math.sqrt((1 + INV_SQRT5) / 2)

# The resulting value simplify a set of expressions. It is much better to compute
# these directly than using trigonometry
A = 0.5  # sin72 * sinAlpha or sin36 * cosAlpha 
B = math.sqrt((2.5 - SQRT5) / 10)  # cos72 * sinAlpha 
C = math.sqrt((2.5 + SQRT5) / 10)  # cos36 * cosAlpha
D = math.sqrt((1 + INV_SQRT5) / 8)  # cos36 * sinAlpha
E = math.sqrt((1 - INV_SQRT5) / 8)  # cos72 * cosAlpha
F = math.sqrt((3 - SQRT5) / 8)  # sin36 * sinAlpha
G = math.sqrt((3 + SQRT5) / 8)  # sin72 * cosAlpha

# Face centers projected onto the z=0 plane & normalized
# 0: North pole,
# 1-5: First pentagon ring
# 6-10: Second pentagon ring
# 11: South pole
face_centers = [
    (0, 0),  # Doesn't actually matter as rotation is 0

    # First ring: five vertices, CCW, multiplied by sinAlpha
    (sin_alpha, 0),  # [cos0, sin0]
    (B, A),  # [cos72, sin72]
    (-D, F),  # [-cos36, sin36]
    (-D, -F),  # [-cos36, -sin36]
    (B, -A),  # [cos72, -sin72]

    # Second ring: the same five vertices but negated (180deg rotation), multiplied by cosAlpha
    (-cos_alpha, 0),  # [-cos0, -sin0]
    (-E, -G),  # [-cos72, -sin72]
    (C, -A),  # [cos36, -sin36]
    (C, A),  # [cos36, sin36]
    (-E, G),  # [-cos72, sin72]

    (0, 0)
]

# Obtain by cross product with the z-axis
axes = [(-y, x) for x, y in face_centers]

# Quaternions are obtained from axis of rotation & angle of rotation
quaternions: List[Tuple[float, float, float, float]] = []
for i, axis in enumerate(axes):
    if i == 0:
        quaternions.append((0, 0, 0, 1))
    elif i == 11:
        quaternions.append((0, -1, 0, 0))  # TODO better to use 1, 0, 0, 0?
    else:
        x, y = axis
        w = cos_alpha if i < 6 else sin_alpha
        quaternions.append((x, y, 0, w))


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import pytest
import json
from pathlib import Path
from a5.projections.authalic import AuthalicProjection
from a5.core.coordinate_systems import Radians
from tests.matchers import is_close

# Load test fixtures
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "authalic.json") as f:
    TEST_DATA = json.load(f)


@pytest.fixture
def authalic():
    return AuthalicProjection()

def test_forward_projections(authalic):
    """Test forward projections match expected values"""
    for test_case in TEST_DATA["forward"]:
        result = authalic.forward(test_case["input"])
        assert is_close(result, test_case["expected"]), \
            f"Expected {test_case['expected']}, got {result}"

def test_round_trip_forward_projections(authalic):
    """Test forward projections can be reversed accurately"""
    for test_case in TEST_DATA["forward"]:
        geodetic = test_case["input"]
        authalic_lat = authalic.forward(geodetic)
        result = authalic.inverse(authalic_lat)
        assert is_close(result, geodetic), \
            f"Expected {geodetic}, got {result}"

def test_inverse_projections(authalic):
    """Test inverse projections match expected values"""
    for test_case in TEST_DATA["inverse"]:
        result = authalic.inverse(test_case["input"])
        assert is_close(result, test_case["expected"]), \
            f"Expected {test_case['expected']}, got {result}"

def test_round_trip_inverse_projections(authalic):
    """Test inverse projections can be reversed accurately"""
    for test_case in TEST_DATA["inverse"]:
        authalic_lat = test_case["input"]
        geodetic = authalic.inverse(authalic_lat)
        result = authalic.forward(geodetic)
        assert is_close(result, authalic_lat), \
            f"Expected {authalic_lat}, got {result}" 


# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import math
from typing import List, Tuple, Optional, TypedDict, NamedTuple, Literal
from .coordinate_systems import Radians, LonLat, Face, Spherical
from ..lattice.types import Orientation
from dataclasses import dataclass

# Type aliases for vectors and matrices (now pure Python)
vec2 = Tuple[float, float]
vec3 = Tuple[float, float, float]
mat2 = Tuple[Tuple[float, float], Tuple[float, float]]
mat2d = Tuple[Tuple[float, float, float], Tuple[float, float, float]]

OriginId = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

class Origin(NamedTuple):
    id: OriginId
    axis: Spherical
    quat: Tuple[float, float, float, float]
    inverse_quat: Tuple[float, float, float, float]
    angle: Radians
    orientation: List[Orientation]
    first_quintant: int

class A5Cell(TypedDict):
    """
    A5 cell with its position information
    """
    origin: Origin  # Origin representing one of pentagon face of the dodecahedron
    segment: int    # Index (0-4) of triangular segment within pentagonal dodecahedron face  
    S: int         # Position along Hilbert curve within triangular segment
    resolution: int # Resolution of the cell


"""
Utility functions for test files - Python equivalent of tests/utils/matchers.ts
"""
from typing import List, Union

# Default tolerance matching TypeScript version
DEFAULT_TOLERANCE = 13


def is_close_array(actual: Union[List, tuple], expected: List, tolerance: int = DEFAULT_TOLERANCE) -> bool:
    """
    Python equivalent of toBeCloseToArray matcher from TypeScript tests.
    Check if arrays are close within specified decimal places tolerance.
    """
    if len(actual) != len(expected):
        return False
    
    tolerance_value = 10**(-tolerance)
    return all(abs(a - e) < tolerance_value for a, e in zip(actual, expected))


def is_close(actual: float, expected: float, tolerance: int = DEFAULT_TOLERANCE) -> bool:
    """Helper function to check if values are close within tolerance"""
    tolerance_value = 10**(-tolerance)
    return abs(actual - expected) < tolerance_value


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from typing import Tuple
from ..core.coordinate_systems import IJ, KJ


def ij_to_kj(ij: IJ) -> KJ:
    """Convert from IJ coordinates to KJ coordinates."""
    i, j = ij
    return (i + j, j)


def kj_to_ij(kj: KJ) -> IJ:
    """Convert from KJ coordinates to IJ coordinates."""
    k, j = kj
    return (k - j, j)



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

# import numpy as np  # Removed
import math
from typing import Tuple, List, Union, Literal, Final
from .coordinate_systems import IJ, KJ

# Type aliases
Quaternary = Literal[0, 1, 2, 3]
YES: Literal[-1] = -1
NO: Literal[1] = 1
Flip = Literal[-1, 1]

class Anchor:
    def __init__(self, k: Quaternary = None, offset: IJ = None, flips: Tuple[Flip, Flip] = None, q: Quaternary = None):
        # Support both old 'k' and new 'q' parameter names
        if q is not None:
            self.q = q
        elif k is not None:
            self.q = k
        else:
            self.q = 0
        self.k = self.q  # Backwards compatibility
        self.offset = offset if offset is not None else (0.0, 0.0)
        self.flips = flips if flips is not None else (NO, NO)

def reverse_pattern(pattern: List[int]) -> List[int]:
    """Reverse a pattern by finding the index of each position."""
    return [pattern.index(i) for i in range(len(pattern))]

# Patterns used to rearrange the cells when shifting
PATTERN = [0, 1, 3, 4, 5, 6, 7, 2]
PATTERN_FLIPPED = [0, 1, 2, 7, 3, 4, 5, 6]
PATTERN_REVERSED = reverse_pattern(PATTERN)
PATTERN_FLIPPED_REVERSED = reverse_pattern(PATTERN_FLIPPED)

# Anchor offset is specified in ij units, the eigenbasis of the Hilbert curve
# Define k as the vector i + j, as it means vectors u & v are of unit length
def ij_to_kj(ij: IJ) -> KJ:
    """Convert from IJ coordinates to KJ coordinates."""
    i, j = ij
    return (i + j, j)

def kj_to_ij(kj: KJ) -> IJ:
    """Convert from KJ coordinates to IJ coordinates."""
    k, j = kj
    return (k - j, j)

#  Orientation of the Hilbert curve. The curve fills a space defined by the triangle with vertices
#  u, v & w. The orientation describes which corner the curve starts and ends at, e.g. wv is a
#  curve that starts at w and ends at v.
Orientation = Literal['uv', 'vu', 'uw', 'wu', 'vw', 'wv']

# Using KJ allows simplification of definitions
k_pos = (1.0, 0.0)  # k
j_pos = (0.0, 1.0)  # j
k_neg = (-k_pos[0], -k_pos[1])
j_neg = (-j_pos[0], -j_pos[1])
ZERO = (0.0, 0.0)

def quaternary_to_kj(n: Quaternary, flips: Tuple[Flip, Flip]) -> KJ:
    """Indirection to allow for flips"""
    flip_x, flip_y = flips
    p = ZERO
    q = ZERO
    
    if flip_x == NO and flip_y == NO:
        p = k_pos
        q = j_pos
    elif flip_x == YES and flip_y == NO:
        # Swap and negate
        p = j_neg
        q = k_neg
    elif flip_x == NO and flip_y == YES:
        # Swap only
        p = j_pos
        q = k_pos
    elif flip_x == YES and flip_y == YES:
        # Negate only
        p = k_neg
        q = j_neg

    if n == 0:
        return ZERO
    elif n == 1:
        return p
    elif n == 2:
        return (p[0] + q[0], p[1] + q[1])
    elif n == 3:
        return (q[0] + 2 * p[0], q[1] + 2 * p[1])
    else:
        raise ValueError(f"Invalid Quaternary value: {n}")

def quaternary_to_flips(n: Quaternary) -> Tuple[Flip, Flip]:
    """Convert quaternary number to flip configuration."""
    flips = [(NO, NO), (NO, YES), (NO, NO), (YES, NO)]
    return flips[n]

FLIP_SHIFT = (-1, 1)

def _shift_digits(digits: List[Quaternary], i: int, flips: List[Flip], invert_j: bool, pattern: List[int]) -> None:
    """Shift digits based on pattern to adjust cell layout."""
    if i <= 0:
        return

    parent_k = digits[i] if i < len(digits) else 0
    child_k = digits[i - 1]
    F = flips[0] + flips[1]

    # Detect when cells need to be shifted
    needs_shift = True
    first = True

    # The value of F which cells need to be shifted
    # The rule is flipped depending on the orientation, specifically on the value of invert_j
    if invert_j != (F == 0):
        needs_shift = parent_k in (1, 2)  # Second & third pentagons only
        first = parent_k == 1  # Second pentagon is first
    else:
        needs_shift = parent_k < 2  # First two pentagons only
        first = parent_k == 0  # First pentagon is first

    if not needs_shift:
        return

    # Apply the pattern by setting the digits based on the value provided
    src = child_k if first else child_k + 4
    dst = pattern[src]
    digits[i - 1] = dst % 4
    digits[i] = (parent_k + 4 + (dst // 4) - (src // 4)) % 4

def s_to_anchor(s: Union[int, str], resolution: int, orientation: Orientation) -> Anchor:
    """Convert s-value to anchor with orientation."""
    input_val = int(s)
    reverse = orientation in ('vu', 'wu', 'vw')
    invert_j = orientation in ('wv', 'vw')
    flip_ij = orientation in ('wu', 'uw')
    
    if reverse:
        input_val = (1 << (2 * resolution)) - input_val - 1
        
    anchor = _s_to_anchor(input_val, resolution, invert_j, flip_ij)
    
    if flip_ij:
        i, j = anchor.offset
        anchor.offset = (j, i)
        
        # Compensate for origin shift
        if anchor.flips[0] == YES:
            anchor.offset = (anchor.offset[0] + FLIP_SHIFT[0], anchor.offset[1] + FLIP_SHIFT[1])
        if anchor.flips[1] == YES:
            anchor.offset = (anchor.offset[0] - FLIP_SHIFT[0], anchor.offset[1] - FLIP_SHIFT[1])
            
    if invert_j:
        i, j = anchor.offset
        new_j = (1 << resolution) - (i + j)
        anchor.flips = (-anchor.flips[0], anchor.flips[1])
        anchor.offset = (anchor.offset[0], new_j)
        
    return anchor

def _s_to_anchor(s: int, resolution: int, invert_j: bool, flip_ij: bool) -> Anchor:
    """Internal function to convert s-value to anchor."""
    offset = [0.0, 0.0]
    flips = [NO, NO]
    
    # Get quaternary digits
    digits = []
    while s > 0 or len(digits) < resolution:
        digits.append(s % 4)
        s >>= 2
    
    # Pad with zeros if needed
    while len(digits) < resolution:
        digits.append(0)
        
    pattern = PATTERN_FLIPPED if flip_ij else PATTERN

    # Process digits from left to right (most significant first)
    for i in range(len(digits) - 1, -1, -1):
        _shift_digits(digits, i, flips, invert_j, pattern)
        new_flips = quaternary_to_flips(digits[i])
        flips[0] *= new_flips[0]
        flips[1] *= new_flips[1]

    flips = [NO, NO]  # Reset flips for the next loop
    for i in range(len(digits) - 1, -1, -1):
        # Scale up existing anchor
        offset = [offset[0] * 2, offset[1] * 2]
        
        # Get child anchor and combine with current anchor
        child_offset = quaternary_to_kj(digits[i], tuple(flips))
        offset = [offset[0] + child_offset[0], offset[1] + child_offset[1]]
        
        new_flips = quaternary_to_flips(digits[i])
        flips[0] *= new_flips[0]
        flips[1] *= new_flips[1]
        
    k = digits[0] if digits else 0

    return Anchor(k, kj_to_ij(tuple(offset)), tuple(flips))

# Get the number of digits needed to represent the offset
# As we don't know the flips we need to add 2 to include the next row
def get_required_digits(offset: Tuple[float, float]) -> int:
    """Calculate required number of digits to represent the offset."""
    index_sum = math.ceil(offset[0]) + math.ceil(offset[1])
    if index_sum == 0:
        return 1
    return 1 + int(math.floor(math.log2(index_sum)))

# This function uses the ij basis, unlike its inverse!
def ij_to_quaternary(ij: IJ, flips: Tuple[Flip, Flip]) -> Quaternary:
    """Convert IJ coordinates to quaternary number with flips."""
    u, v = ij
    digit = 0
    
    # Boundaries to compare against
    a = -(u + v) if flips[0] == YES else u + v
    b = -u if flips[1] == YES else u
    c = -v if flips[0] == YES else v
    
    # Only one flip
    if flips[0] + flips[1] == 0:
        if c < 1:
            digit = 0
        elif b > 1:
            digit = 3
        elif a > 1:
            digit = 2
        else:
            digit = 1
    # No flips or both
    else:
        if a < 1:
            digit = 0
        elif b > 1:
            digit = 3
        elif c > 1:
            digit = 2
        else:
            digit = 1
            
    return digit

def ij_to_s(input_ij: IJ, resolution: int, orientation: str = 'uv') -> int:
    """Convert IJ coordinates to s-value with orientation."""
    reverse = orientation in ('vu', 'wu', 'vw')
    invert_j = orientation in ('wv', 'vw')
    flip_ij = orientation in ('wu', 'uw')
    
    # Convert tuple to list for modification, then back to tuple
    ij = list(input_ij)
    if flip_ij:
        ij[0], ij[1] = ij[1], ij[0]
    if invert_j:
        i, j = ij
        ij[1] = (1 << resolution) - (i + j)
    
    ij = tuple(ij)  # Convert back to tuple
        
    s = _ij_to_s(ij, invert_j, flip_ij, resolution)
    if reverse:
        s = (1 << (2 * resolution)) - s - 1
        
    return s

def _ij_to_s(input_ij: IJ, invert_j: bool, flip_ij: bool, resolution: int) -> int:
    """Internal function to convert IJ coordinates to s-value."""
    # Get number of digits we need to process
    num_digits = resolution
    digits = [0] * num_digits
    
    flips = [NO, NO]
    pivot = [0.0, 0.0]
    
    # Process digits from left to right (most significant first)
    for i in range(num_digits - 1, -1, -1):
        relative_offset = (input_ij[0] - pivot[0], input_ij[1] - pivot[1])
        scale = 1 << i
        scaled_offset = (relative_offset[0] / scale, relative_offset[1] / scale)
        
        digit = ij_to_quaternary(scaled_offset, tuple(flips))
        digits[i] = digit
        
        # Update running state
        child_offset = kj_to_ij(quaternary_to_kj(digit, tuple(flips)))
        upscaled_child_offset = (child_offset[0] * scale, child_offset[1] * scale)
        pivot = [pivot[0] + upscaled_child_offset[0], pivot[1] + upscaled_child_offset[1]]
        
        new_flips = quaternary_to_flips(digit)
        flips[0] *= new_flips[0]
        flips[1] *= new_flips[1]
        
    pattern = PATTERN_FLIPPED_REVERSED if flip_ij else PATTERN_REVERSED

    for i in range(num_digits):
        new_flips = quaternary_to_flips(digits[i])
        flips[0] *= new_flips[0]
        flips[1] *= new_flips[1]
        _shift_digits(digits, i, flips, invert_j, pattern)
        
    # Convert digits to s-value
    output = 0
    for i in range(num_digits - 1, -1, -1):
        scale = 1 << (2 * i)
        output += digits[i] * scale
        
    return output


# Adaptation of icoVertexGreatCircle.ec from DGGAL project
# BSD 3-Clause License
# 
# Copyright (c) 2014-2025, Ecere Corporation
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# BSD 3-Clause License
# Copyright (c) 2024, A5 Project Contributors
# All rights reserved.

"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import cast, Dict, Tuple
from ..core.coordinate_systems import Cartesian, Face, Barycentric, FaceTriangle, SphericalTriangle
from ..core.coordinate_transforms import face_to_barycentric, barycentric_to_face
from ..geometry.spherical_triangle import SphericalTriangleShape
from ..math import vec3, quat

class PolyhedralProjection:
    """
    Polyhedral Equal Area projection using Slice & Dice algorithm
    """
    
    def __init__(self):
        # Cache for triangle-dependent calculations in inverse projection
        self._inverse_triangle_cache: Dict[Tuple, Dict] = {}
    
    def forward(self, v: Cartesian, spherical_triangle: SphericalTriangle, face_triangle: FaceTriangle) -> Face:
        """
        Forward projection: converts a spherical point to face coordinates
        
        Args:
            v: The spherical point to project
            spherical_triangle: The spherical triangle vertices
            face_triangle: The face triangle vertices
            
        Returns:
            The face coordinates
        """
        A, B, C = spherical_triangle
        triangle_shape = SphericalTriangleShape(spherical_triangle)

        # When v is close to A, the quadruple product is unstable.
        # As we just need the intersection of two great circles we can use difference
        # between A and v, as it lies in the same plane of the great circle containing A & v
        Z = vec3.create()
        vec3.subtract(Z, v, A)
        vec3.normalize(Z, Z)
        Z = cast(Cartesian, (Z[0], Z[1], Z[2]))
        
        p = vec3.create()
        vec3.quadrupleProduct(p, A, Z, B, C)
        vec3.normalize(p, p)
        p = cast(Cartesian, (p[0], p[1], p[2]))

        h = vec3.vectorDifference(A, v) / vec3.vectorDifference(A, p)
        Area_ABC = triangle_shape.get_area()
        scaled_area = h / Area_ABC
        
        b = cast(Barycentric, (
            1 - h,
            scaled_area * SphericalTriangleShape([A, p, C]).get_area(),
            scaled_area * SphericalTriangleShape([A, B, p]).get_area()
        ))
        
        return barycentric_to_face(b, face_triangle)

    def inverse(self, face_point: Face, face_triangle: FaceTriangle, spherical_triangle: SphericalTriangle) -> Cartesian:
        """
        Inverse projection: converts face coordinates back to spherical coordinates
        
        Args:
            face_point: The face coordinates
            face_triangle: The face triangle vertices
            spherical_triangle: The spherical triangle vertices
            
        Returns:
            The spherical coordinates
        """
        A, B, C = spherical_triangle
        b = face_to_barycentric(face_point, face_triangle)

        threshold = 1 - 1e-14
        if b[0] > threshold:
            return A
        if b[1] > threshold:
            return B
        if b[2] > threshold:
            return C
        
        # Get cached triangle-dependent constants
        constants = self._get_triangle_constants(spherical_triangle)
        area_abc = constants['area_abc']
        c1 = constants['c1']
        c01 = constants['c01']
        c12 = constants['c12']
        c20 = constants['c20']
        s12 = constants['s12']
        V = constants['V']
        
        # Point-dependent calculations
        h = 1 - b[0]
        R = b[2] / h
        alpha = R * area_abc
        S = math.sin(alpha)
        half_c = math.sin(alpha / 2)
        CC = 2 * half_c * half_c  # Half angle formula

        f = S * V + CC * (c01 * c12 - c20)
        g = CC * s12 * (1 + c01)
        q = (2 / math.acos(c12)) * math.atan2(g, f)
        
        # Use gl-matrix style slerp for P = slerp(B, C, q)
        P = vec3.create()
        vec3.slerp(P, B, C, q)
        P = cast(Cartesian, (P[0], P[1], P[2]))
        
        # K = A - P  
        K = vec3.vectorDifference(A, P)
        t = self._safe_acos(h * K) / self._safe_acos(K)
        
        # Final slerp: out = slerp(A, P, t)
        out = [0.0, 0.0, 0.0]
        vec3.slerp(out, A, P, t)
        return cast(Cartesian, (out[0], out[1], out[2]))

    def _get_triangle_constants(self, spherical_triangle: SphericalTriangle):
        """
        Get cached triangle-dependent constants for inverse projection.
        These values only depend on the spherical triangle, not the input point.
        """
        # Create a cache key from the triangle vertices
        # Convert to tuples since lists aren't hashable
        A, B, C = spherical_triangle
        cache_key = (tuple(A), tuple(B), tuple(C))
        
        if cache_key not in self._inverse_triangle_cache:
            triangle_shape = SphericalTriangleShape(spherical_triangle)
            c1 = vec3.create()
            vec3.cross(c1, B, C)
            
            constants = {
                'area_abc': triangle_shape.get_area(),
                'c1': (c1[0], c1[1], c1[2]),  # Store as tuple
                'c01': vec3.dot(A, B),
                'c12': vec3.dot(B, C),
                'c20': vec3.dot(C, A),
                's12': vec3.length(c1),
                'V': vec3.dot(A, c1)  # Triple product of A, B, C
            }
            self._inverse_triangle_cache[cache_key] = constants
        
        return self._inverse_triangle_cache[cache_key]

    def _safe_acos(self, x: float) -> float:
        """
        Computes acos(1 - 2 * x * x) without loss of precision for small x
        
        Args:
            x: Input value
            
        Returns:
            acos(1 - x)
        """
        if x < 1e-3:
            return (2 * x + x * x * x / 3)
        else:
            return math.acos(1 - 2 * x * x) 


"""
Tests for origin-related functionality.
"""

import pytest
import json
import math
from pathlib import Path
from tests.matchers import is_close_array

from a5.core.origin import (    
    origins,
    find_nearest_origin,
    haversine,
    quintant_to_segment,
    segment_to_quintant
)
from a5.core.constants import PI_OVER_5
from a5.core.coordinate_transforms import to_cartesian
from a5.math.vec3 import length
from a5.math.quat import length as quat_length

# Load test fixtures
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "origins.json") as f:
    EXPECTED_ORIGINS = json.load(f)

def test_origin_constants():
    """Test that we have 12 origins for dodecahedron faces."""
    assert len(origins) == 12


def test_should_match_expected_origins_from_json_file():
    """Test that origins match expected values from JSON fixture"""
    assert len(origins) == len(EXPECTED_ORIGINS)
    for i, origin in enumerate(origins):
        expected = EXPECTED_ORIGINS[i]
        assert origin.id == expected['id']
        assert origin.angle == expected['angle']
        assert origin.orientation == expected['orientation']
        assert origin.first_quintant == expected['firstQuintant']
        assert is_close_array(origin.axis, expected['axis']), \
            f"Origin {i} axis wrong: expected {expected['axis']}, got {origin.axis}"
        assert is_close_array(origin.quat, expected['quat']), \
            f"Origin {i} quat wrong: expected {expected['quat']}, got {origin.quat}"

def test_origin_properties():
    """Test that each origin has required properties."""
    for origin in origins:
        # Check properties exist
        assert origin.axis is not None
        assert origin.quat is not None
        assert origin.angle is not None
        
        # Check axis is unit vector when converted to cartesian
        cartesian = to_cartesian(origin.axis)
        vector_length = length(cartesian)
        assert abs(vector_length - 1.0) < 1e-15
        
        # Check quaternion is normalized
        q_length = quat_length(origin.quat)
        assert abs(q_length - 1.0) < 1e-10


def test_find_nearest_origin():
    """Test finding nearest origin for various points."""
    # Test points at face centers
    for origin in origins:
        point = origin.axis
        nearest = find_nearest_origin(point)
        assert nearest == origin

    # Test points at face boundaries
    boundary_points = [
        # Between north pole and equatorial faces
        {"point": [0, PI_OVER_5/2], "expected_origins": [0, 1]},
        # Between equatorial faces
        {"point": [2*PI_OVER_5, PI_OVER_5], "expected_origins": [3, 4]},
        # Between equatorial and south pole
        {"point": [0, math.pi - PI_OVER_5/2], "expected_origins": [9, 10]},
    ]

    for test_case in boundary_points:
        nearest = find_nearest_origin(test_case["point"])
        assert nearest.id in test_case["expected_origins"]

    # Test antipodal points
    for origin in origins:
        theta, phi = origin.axis
        # Add π to theta and π-phi to get antipodal point
        antipodal = [theta + math.pi, math.pi - phi]
        
        nearest = find_nearest_origin(antipodal)
        # Should find one of the faces near the antipodal point
        assert nearest != origin


def test_haversine():
    """Test haversine distance calculations."""
    # Test identical points
    point = [0, 0]
    assert haversine(point, point) == 0

    point2 = [math.pi/4, math.pi/3]
    assert haversine(point2, point2) == 0

    # Test symmetry
    p1 = [0, math.pi/4]
    p2 = [math.pi/2, math.pi/3]
    
    d1 = haversine(p1, p2)
    d2 = haversine(p2, p1)
    
    assert abs(d1 - d2) < 1e-15

    # Test increasing distance
    origin = [0, 0]
    distances = [
        [0, math.pi/6],      # 30°
        [0, math.pi/4],      # 45°
        [0, math.pi/3],      # 60°
        [0, math.pi/2],      # 90°
    ]

    last_distance = 0
    for point in distances:
        distance = haversine(origin, point)
        assert distance > last_distance
        last_distance = distance

    # Test longitude separation
    lat = math.pi/4  # Fixed latitude
    p1 = [0, lat]
    p2 = [math.pi, lat]
    p3 = [math.pi/2, lat]

    d1 = haversine(p1, p2)  # 180° separation
    d2 = haversine(p1, p3)  # 90° separation

    assert d1 > d2

    # Test known cases
    test_cases = [
        {
            "p1": [0, 0],
            "p2": [0, math.pi/2],
            "expected": 0.5  # sin²(π/4) = 0.5
        },
        {
            "p1": [0, math.pi/4],
            "p2": [math.pi/2, math.pi/4],
            "expected": 0.25  # For points at same latitude
        }
    ]

    for case in test_cases:
        assert abs(haversine(case["p1"], case["p2"]) - case["expected"]) < 1e-4

def test_quintant_conversion():
    """Test conversion between quintants and segments."""
    origin = origins[0]
    for quintant in range(5):
        segment, orientation = quintant_to_segment(quintant, origin)
        round_trip_quintant, round_trip_orientation = segment_to_quintant(segment, origin)
        assert round_trip_quintant == quintant 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import json
from pathlib import Path
from a5.traversal.grid_disk import grid_disk, grid_disk_vertex
from a5.core.serialization import get_resolution
from a5.core.compact import uncompact


def load_fixtures():
    fixture_path = Path(__file__).parent / "fixtures" / "grid-disk.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


def hex_to_int(hex_str: str) -> int:
    return int(hex_str, 16)


class TestGridDisk:
    def test_grid_disk_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures:
            cell_id = hex_to_int(case["cellId"])
            k = case["k"]
            target_res = get_resolution(cell_id)
            expected = sorted(hex_to_int(h) for h in case["cells"])
            result = sorted(uncompact(grid_disk(cell_id, k), target_res))
            assert result == expected, \
                f'cellId={case["cellId"]}, k={k}: got {len(result)} cells, expected {len(expected)}'


class TestGridDiskVertex:
    def test_grid_disk_vertex_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures:
            cell_id = hex_to_int(case["cellId"])
            k = case["k"]
            target_res = get_resolution(cell_id)
            # grid_disk_vertex returns edge + vertex cells
            extra = [hex_to_int(h) for h in case.get("extraVertexCells", [])]
            expected_edge = [hex_to_int(h) for h in case["cells"]]
            expected = sorted(set(expected_edge + extra))
            result = sorted(uncompact(grid_disk_vertex(cell_id, k), target_res))
            assert result == expected, \
                f'cellId={case["cellId"]}, k={k}: got {len(result)} cells, expected {len(expected)}'



"""
Tests for cell-info functionality.
"""

import json
import pytest
from pathlib import Path
from a5.core.cell_info import get_num_cells, cell_area

# Load test fixtures  
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "cell-info.json") as f:
    CELL_INFO_FIXTURES = json.load(f)

def test_get_num_cells_returns_correct_count_for_all_resolutions():
    """Test that getNumCells returns correct number of cells for all resolutions."""
    for fixture in CELL_INFO_FIXTURES["numCells"]:
        result = get_num_cells(fixture["resolution"])
        expected = int(fixture["countBigInt"])  # Use the exact BigInt value
        assert result == expected, f"Resolution {fixture['resolution']}: got {result}, expected {expected}"

def test_cell_area_returns_correct_area_for_all_resolutions():
    """Test that cellArea returns correct area for all resolutions."""  
    for fixture in CELL_INFO_FIXTURES["cellArea"]:
        assert cell_area(fixture["resolution"]) == fixture["areaM2"]


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import math
from typing import List, Tuple
from ..geometry.pentagon import PentagonShape, Pentagon
from .pentagon import a, BASIS, PENTAGON, TRIANGLE, v, V, w
from .constants import TWO_PI, TWO_PI_OVER_5
from ..lattice import NO, Anchor, YES
from ..math import vec2

TRIANGLE_MODE = False

shift_right = w  # No need to copy, just reference
shift_left = (-w[0], -w[1])

# Define transforms for each pentagon in the primitive unit
# Using pentagon vertices and angle as the basis for the transform
QUINTANT_ROTATIONS = [
    (
        (math.cos(TWO_PI_OVER_5 * quintant), -math.sin(TWO_PI_OVER_5 * quintant)),
        (math.sin(TWO_PI_OVER_5 * quintant), math.cos(TWO_PI_OVER_5 * quintant))
    )
    for quintant in range(5)
]

def get_pentagon_vertices(resolution: int, quintant: int, anchor: Anchor) -> PentagonShape:
    """
    Get pentagon vertices
    
    Args:
        resolution: The resolution level
        quintant: The quintant index (0-4)
        anchor: The anchor information
        
    Returns:
        A pentagon shape with transformed vertices
    """
    pentagon = (TRIANGLE if TRIANGLE_MODE else PENTAGON).clone()
    
    # Matrix-vector multiplication using gl-matrix style: BASIS @ anchor.offset
    # Convert 2x2 matrix from ((a,b),(c,d)) to [a,c,b,d] (column-major)
    basis_flat = [BASIS[0][0], BASIS[1][0], BASIS[0][1], BASIS[1][1]]
    translation_vec = vec2.create()
    vec2.transformMat2(translation_vec, anchor.offset, basis_flat)
    translation = (translation_vec[0], translation_vec[1])

    # Apply transformations based on anchor properties
    if anchor.flips[0] == NO and anchor.flips[1] == YES:
        pentagon.rotate180()

    q = anchor.q
    F = anchor.flips[0] + anchor.flips[1]
    if (
        # Orient last two pentagons when both or neither flips are YES
        ((F == -2 or F == 2) and q > 1) or
        # Orient first & last pentagons when only one of flips is YES
        (F == 0 and (q == 0 or q == 3))
    ):
        pentagon.reflect_y()

    if anchor.flips[0] == YES and anchor.flips[1] == YES:
        pentagon.rotate180()
    elif anchor.flips[0] == YES:
        pentagon.translate(shift_left)
    elif anchor.flips[1] == YES:
        pentagon.translate(shift_right)

    # Position within quintant
    pentagon.translate(translation)
    pentagon.scale(1 / (2 ** resolution))
    pentagon.transform(QUINTANT_ROTATIONS[quintant])

    return pentagon

PentagonFlavor = int  # 0-7


def get_pentagon_flavor(anchor: Anchor) -> PentagonFlavor:
    """Get the flavor (0-7) of a pentagon from its anchor."""
    f = 0
    if anchor.flips[1] == YES:
        f += 2

    q = anchor.q
    F = anchor.flips[0] + anchor.flips[1]
    if (
        # Orient last two pentagons when both or neither flips are YES
        ((F == -2 or F == 2) and q > 1) or
        # Orient first & last pentagons when only one of flips is YES
        (F == 0 and (q == 0 or q == 3))
    ):
        f += 1

    if F == -2 or F == 2:
        f += 4

    return f


def get_quintant_vertices(quintant: int) -> PentagonShape:
    triangle = TRIANGLE.clone()
    triangle.transform(QUINTANT_ROTATIONS[quintant])
    return triangle

def get_face_vertices() -> PentagonShape:
    vertices = []
    for rotation in QUINTANT_ROTATIONS:
        # Matrix-vector multiplication using gl-matrix style: rotation @ v
        # Convert 2x2 matrix from ((a,b),(c,d)) to [a,c,b,d] (column-major)
        rotation_flat = [rotation[0][0], rotation[1][0], rotation[0][1], rotation[1][1]]
        vertex_vec = vec2.create()
        vec2.transformMat2(vertex_vec, v, rotation_flat)
        new_vertex = (vertex_vec[0], vertex_vec[1])
        vertices.append(new_vertex)
    return PentagonShape(vertices)

def get_quintant_polar(polar: Tuple[float, float]) -> int:
    """
    Determines which quintant a polar coordinate belongs to.
    
    Args:
        polar: Polar coordinates (r, theta)
        
    Returns:
        Quintant index (0-4)
    """
    rho, gamma = polar
    return (round(gamma / TWO_PI_OVER_5) + 5) % 5


"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import cast, List
from .coordinate_systems import (
    Degrees, Radians, Face, Polar, IJ, Cartesian, Spherical, LonLat,
    Barycentric, FaceTriangle
)
from .pentagon import BASIS_INVERSE, BASIS
from ..projections.authalic import AuthalicProjection
from ..math import vec2, vec3

# Create singleton instance like TypeScript
authalic = AuthalicProjection()

# Constants
LONGITUDE_OFFSET = cast(Degrees, 93.0)  # degrees

def deg_to_rad(deg: Degrees) -> Radians:
    """Convert degrees to radians."""
    return cast(Radians, deg * (math.pi / 180))

def rad_to_deg(rad: Radians) -> Degrees:
    """Convert radians to degrees."""
    return cast(Degrees, rad * (180 / math.pi))

def to_polar(xy: Face) -> Polar:
    """Convert face coordinates to polar coordinates."""
    rho = vec2.length(xy)  # Radial distance from face center
    gamma = cast(Radians, math.atan2(xy[1], xy[0]))  # Azimuthal angle
    return cast(Polar, (rho, gamma))

def to_face(polar: Polar) -> Face:
    """Convert polar coordinates to face coordinates."""
    rho, gamma = polar
    x = rho * math.cos(gamma)
    y = rho * math.sin(gamma)
    return cast(Face, (x, y))

def face_to_ij(face: Face) -> IJ:
    """Convert face coordinates to IJ coordinates."""
    # Use gl-matrix style transformation
    # Convert 2x2 matrix from ((a,b),(c,d)) to [a,c,b,d] (column-major)
    basis_flat = [BASIS_INVERSE[0][0], BASIS_INVERSE[1][0], BASIS_INVERSE[0][1], BASIS_INVERSE[1][1]]
    out = vec2.create()
    vec2.transformMat2(out, face, basis_flat)
    return cast(IJ, (out[0], out[1]))

def ij_to_face(ij: IJ) -> Face:
    """Convert IJ coordinates to face coordinates."""
    # Use gl-matrix style transformation
    # Convert 2x2 matrix from ((a,b),(c,d)) to [a,c,b,d] (column-major)
    basis_flat = [BASIS[0][0], BASIS[1][0], BASIS[0][1], BASIS[1][1]]
    out = vec2.create()
    vec2.transformMat2(out, ij, basis_flat)
    return cast(Face, (out[0], out[1]))

def to_spherical(xyz: Cartesian) -> Spherical:
    """Convert Cartesian coordinates to spherical coordinates."""
    theta = cast(Radians, math.atan2(xyz[1], xyz[0]))
    r = vec3.length(xyz)
    phi = cast(Radians, math.acos(xyz[2] / r))
    return cast(Spherical, (theta, phi))

def to_cartesian(spherical: Spherical) -> Cartesian:
    """Convert spherical coordinates to Cartesian coordinates."""
    theta, phi = spherical
    x = math.sin(phi) * math.cos(theta)
    y = math.sin(phi) * math.sin(theta)
    z = math.cos(phi)
    return cast(Cartesian, (x, y, z))

def from_lonlat(lon_lat: LonLat) -> Spherical:
    """Convert longitude/latitude to spherical coordinates.
    
    Args:
        lon_lat: Tuple of (longitude, latitude) in degrees
            longitude: 0 to 360
            latitude: -90 to 90
    
    Returns:
        Tuple of (theta, phi) in radians
    """
    longitude, latitude = lon_lat
    theta = deg_to_rad(cast(Degrees, longitude + LONGITUDE_OFFSET))
    
    geodetic_lat = deg_to_rad(cast(Degrees, latitude))
    authalic_lat = authalic.forward(geodetic_lat)
    phi = cast(Radians, math.pi / 2 - authalic_lat)
    return cast(Spherical, (theta, phi))

def to_lonlat(spherical: Spherical) -> LonLat:
    """Convert spherical coordinates to longitude/latitude.
    
    Args:
        spherical: Tuple of (theta, phi) in radians
            theta: 0 to 2π
            phi: 0 to π
    
    Returns:
        Tuple of (longitude, latitude) in degrees
            longitude: 0 to 360
            latitude: -90 to 90
    """
    theta, phi = spherical
    longitude = rad_to_deg(theta) - LONGITUDE_OFFSET

    authalic_lat = cast(Radians, math.pi / 2 - phi)
    geodetic_lat = authalic.inverse(authalic_lat)
    latitude = rad_to_deg(geodetic_lat)
    return cast(LonLat, (longitude, latitude))

def face_to_barycentric(p: Face, triangle: FaceTriangle) -> Barycentric:
    """Convert face coordinates to barycentric coordinates."""
    p1, p2, p3 = triangle
    d31 = [p1[0] - p3[0], p1[1] - p3[1]]
    d23 = [p3[0] - p2[0], p3[1] - p2[1]]
    d3p = [p[0] - p3[0], p[1] - p3[1]]
    
    det = d23[0] * d31[1] - d23[1] * d31[0]
    b0 = (d23[0] * d3p[1] - d23[1] * d3p[0]) / det
    b1 = (d31[0] * d3p[1] - d31[1] * d3p[0]) / det
    b2 = 1 - (b0 + b1)
    return cast(Barycentric, (b0, b1, b2))

def barycentric_to_face(b: Barycentric, triangle: FaceTriangle) -> Face:
    """Convert barycentric coordinates to face coordinates."""
    p1, p2, p3 = triangle
    return cast(Face, (
        b[0] * p1[0] + b[1] * p2[0] + b[2] * p3[0],
        b[0] * p1[1] + b[1] * p2[1] + b[2] * p3[1]
    ))

Contour = List[LonLat]

def normalize_longitudes(contour: Contour) -> Contour:
    """Normalizes longitude values in a contour to handle antimeridian crossing.
    
    Args:
        contour: Array of [longitude, latitude] points
        
    Returns:
        Normalized contour with consistent longitude values
    """
    # Calculate center in Cartesian space to avoid poles & antimeridian crossing issues
    points = [to_cartesian(from_lonlat(lonlat)) for lonlat in contour]
    center = vec3.create()
    for point in points:
        vec3.add(center, center, point)
    
    # Normalize the center
    vec3.normalize(center, center)
    center_lon, center_lat = to_lonlat(to_spherical(cast(Cartesian, (center[0], center[1], center[2]))))
    
    if center_lat > 89.99 or center_lat < -89.99:
        # Near poles, use first point's longitude
        center_lon = contour[0][0]

    # Normalize center longitude to be in the range -180 to 180
    center_lon = ((center_lon + 180) % 360 + 360) % 360 - 180

    # Normalize each point relative to center
    result = []
    for point in contour:
        longitude, latitude = point
        
        # Adjust longitude to be closer to center
        while longitude - center_lon > 180:
            longitude = cast(Degrees, longitude - 360)
        while longitude - center_lon < -180:
            longitude = cast(Degrees, longitude + 360)
        result.append(cast(LonLat, (longitude, latitude)))
    
    return result 


import pytest
import math
from a5.core.hilbert import (
    quaternary_to_kj,
    quaternary_to_flips,
    YES,
    NO,
    Anchor,
    ij_to_kj,
    kj_to_ij,
    get_required_digits,
    ij_to_s,
    s_to_anchor,
)

def test_hilbert_anchor_base_cases():
    # Test first corner (0)
    offset0 = quaternary_to_kj(0, (NO, NO))
    assert offset0 == (0, 0)
    flips0 = quaternary_to_flips(0)
    assert flips0 == (NO, NO)

    # Test second corner (1)
    offset1 = quaternary_to_kj(1, (NO, NO))
    assert offset1 == (1, 0)
    flips1 = quaternary_to_flips(1)
    assert flips1 == (NO, YES)

    # Test third corner (2)
    offset2 = quaternary_to_kj(2, (NO, NO))
    assert offset2 == (1, 1)
    flips2 = quaternary_to_flips(2)
    assert flips2 == (NO, NO)

    # Test fourth corner (3)
    offset3 = quaternary_to_kj(3, (NO, NO))
    assert offset3 == (2, 1)
    flips3 = quaternary_to_flips(3)
    assert flips3 == (YES, NO)

def test_hilbert_anchor_respects_flips():
    # Test with x-flip
    offset_x = quaternary_to_kj(1, (YES, NO))
    assert offset_x == (0, -1)

    # Test with y-flip
    offset_y = quaternary_to_kj(1, (NO, YES))
    assert offset_y == (0, 1)

    # Test with both flips
    offset_xy = quaternary_to_kj(1, (YES, YES))
    assert offset_xy == (-1, 0)

def test_output_flips_depend_only_on_input():
    EXPECTED_FLIPS = [
        (NO, NO),
        (NO, YES),
        (NO, NO),
        (YES, NO)
    ]
    for n in range(4):
        flips = quaternary_to_flips(n)
        assert flips == EXPECTED_FLIPS[n]

def test_generates_correct_sequence():
    # Test first few indices
    anchor0 = s_to_anchor(0, 1, 'uv')
    assert tuple(anchor0.offset) == (0, 0)
    assert anchor0.flips == (NO, NO)

    anchor1 = s_to_anchor(1, 1, 'uv')
    assert anchor1.flips[1] == YES

    anchor4 = s_to_anchor(4, 1, 'uv')
    assert math.sqrt(anchor4.offset[0]**2 + anchor4.offset[1]**2) > 1  # Should be scaled up

    # Test that sequence length grows exponentially
    anchors = [s_to_anchor(i, 1, 'uv') for i in range(16)]
    unique_offsets = {tuple(a.offset) for a in anchors}
    assert len(unique_offsets) == 12
    unique_anchors = {(tuple(a.offset), a.flips) for a in anchors}
    assert len(unique_anchors) == 15

def test_neighboring_anchors_are_adjacent():
    # Test that combining anchors preserves orientation rules
    anchor1 = s_to_anchor(0, 1, 'uv')
    anchor2 = s_to_anchor(1, 1, 'uv')
    anchor3 = s_to_anchor(2, 1, 'uv')
    
    # Check that relative positions make sense
    diff = (anchor2.offset[0] - anchor1.offset[0], anchor2.offset[1] - anchor1.offset[1])
    assert math.sqrt(diff[0]**2 + diff[1]**2) == 1  # Should be adjacent
    diff2 = (anchor3.offset[0] - anchor2.offset[0], anchor3.offset[1] - anchor2.offset[1])
    assert math.sqrt(diff2[0]**2 + diff2[1]**2) == math.sqrt(2)  # Should be adjacent

def test_generates_correct_anchors():
    EXPECTED_ANCHORS = [
        {'s': 0, 'offset': [0, 0], 'flips': (NO, NO)},
        {'s': 9, 'offset': [3, 1], 'flips': (YES, YES)},
        {'s': 16, 'offset': [2, 2], 'flips': (NO, NO)},
        {'s': 17, 'offset': [3, 2], 'flips': (NO, YES)},
        {'s': 31, 'offset': [1, 3], 'flips': (YES, NO)},
        {'s': 77, 'offset': [7, 5], 'flips': (NO, NO)},
        {'s': 100, 'offset': [3, 7], 'flips': (YES, YES)},
        {'s': 101, 'offset': [2, 7], 'flips': (YES, NO)},
        {'s': 170, 'offset': [10, 1], 'flips': (NO, NO)},
        {'s': 411, 'offset': [7, 13], 'flips': (YES, NO)},
        {'s': 1762, 'offset': [7, 31], 'flips': (YES, NO)},
        {'s': 481952, 'offset': [96, 356], 'flips': (YES, YES)},
        {'s': 192885192, 'offset': [13183, 1043], 'flips': (NO, NO)},
        {'s': 4719283155, 'offset': [37190, 46076], 'flips': (NO, YES)},
        {'s': 7123456789, 'offset': [29822, 40293], 'flips': (NO, YES)},
    ]

    for test_case in EXPECTED_ANCHORS:
        anchor = s_to_anchor(test_case['s'], 20, 'uv')
        assert list(anchor.offset) == test_case['offset']
        assert anchor.flips == test_case['flips']

def test_ij_to_kj_conversion():
    # Test some basic conversions
    test_cases = [
        ([0, 0], [0, 0]),    # Origin
        ([1, 0], [1, 0]),    # Unit i
        ([0, 1], [1, 1]),    # Unit j -> k=i+j=1, j=1
        ([1, 1], [2, 1]),    # i + j -> k=2, j=1
        ([2, 3], [5, 3])     # 2i + 3j -> k=5, j=3
    ]

    for input_ij, expected_kj in test_cases:
        result = ij_to_kj(input_ij)
        assert list(result) == expected_kj

def test_kj_to_ij_conversion():
    # Test some basic conversions
    test_cases = [
        ([0, 0], [0, 0]),     # Origin
        ([1, 0], [1, 0]),     # Pure k -> i=1, j=0
        ([1, 1], [0, 1]),     # k=1, j=1 -> i=0, j=1
        ([2, 1], [1, 1]),     # k=2, j=1 -> i=1, j=1
        ([5, 3], [2, 3])      # k=5, j=3 -> i=2, j=3
    ]

    for input_kj, expected_ij in test_cases:
        result = kj_to_ij(input_kj)
        assert list(result) == expected_ij

def test_ij_kj_inverse():
    # Test that converting back and forth gives the original coordinates
    test_points = [
        [0, 0],
        [1, 0],
        [0, 1],
        [1, 1],
        [2, 3],
        [-1, 2],
        [3, -2]
    ]

    for point in test_points:
        kj = ij_to_kj(point)
        ij = kj_to_ij(kj)
        assert list(ij) == point

def test_get_required_digits():
    test_cases = [
        ([0, 0], 1),
        ([1, 0], 1),
        ([2, 1], 2),
        ([4, 0], 3),
        ([8, 8], 5),
        ([16, 0], 5),
        ([32, 32], 7)
    ]

    for offset, expected in test_cases:
        assert get_required_digits(offset) == expected

def test_required_digits_matches_s_to_anchor():
    # Test that get_required_digits matches the number of digits 
    # actually used in sToAnchor's output
    test_values = [0, 1, 2, 3, 4, 9, 16, 17, 31, 77, 100]
    
    for s in test_values:
        anchor = s_to_anchor(s, 20, 'uv')
        required_digits = get_required_digits(anchor.offset)
        # Convert to base 4 manually
        temp_s = s
        digits = 0
        while temp_s > 0:
            digits += 1
            temp_s //= 4
        actual_digits = digits if digits > 0 else 1
        assert required_digits >= actual_digits
        assert required_digits <= actual_digits + 1

def test_ij_to_s():
    test_values = [
        # First quadrant
        {'s': 0, 'offset': [0, 0]},
        {'s': 0, 'offset': [0.999, 0]},
        {'s': 1, 'offset': [0.6, 0.6]},
        {'s': 7, 'offset': [0.000001, 1.1]},
        {'s': 2, 'offset': [1.2, 0.5]},
        {'s': 2, 'offset': [1.9999, 0]},

        # Recursive cases, 2nd quadrant, flipY
        {'s': 3, 'offset': [1.9999, 0.001]},
        {'s': 4, 'offset': [1.1, 1.1]},
        {'s': 5, 'offset': [1.999, 1.999]},
        {'s': 6, 'offset': [0.99, 1.99]},

        # 3rd quadrant, no flips
        {'s': 28, 'offset': [0.999, 2.000001]},
        {'s': 29, 'offset': [0.9, 2.5]},
        {'s': 30, 'offset': [0.5, 3.1]},
        {'s': 31, 'offset': [1.3, 2.5]},

        # 4th quadrant, flipX
        {'s': 8, 'offset': [2.00001, 1.001]},
        {'s': 9, 'offset': [2.8, 0.5]},
        {'s': 10, 'offset': [2.00001, 0.5]},
        {'s': 11, 'offset': [3.5, 0.2]},

        # Next level
        {'s': 15, 'offset': [2.5, 1.5]},
        {'s': 21, 'offset': [3.999, 3.999]},

        # Both flips
        {'s': 24, 'offset': [1.999, 3.999]},
        {'s': 25, 'offset': [1.2, 3.5]},
        {'s': 26, 'offset': [1.9, 2.2]},
        {'s': 27, 'offset': [0.1, 3.9]}
    ]

    for test_case in test_values:
        result = ij_to_s(test_case['offset'], 3, 'uv')
        assert result == test_case['s']

@pytest.mark.parametrize('orientation', ['uv', 'vu', 'uw', 'wu', 'vw', 'wv'])
@pytest.mark.parametrize('s', [0, 1, 2, 3, 4, 9, 16, 17, 31, 77, 100, 101, 170, 411, 1762, 4410, 12387, 41872])
def test_ij_to_s_inverse_of_s_to_anchor(s, orientation):
    resolution = 20
    anchor = s_to_anchor(s, resolution, orientation)

    # Nudge the offset away from the edge of the triangle
    flip_x, flip_y = anchor.flips
    if flip_x == NO and flip_y == NO:
        anchor.offset = [anchor.offset[0] + 0.1, anchor.offset[1] + 0.1]
    elif flip_x == YES and flip_y == NO:
        anchor.offset = [anchor.offset[0] + 0.1, anchor.offset[1] - 0.2]
    elif flip_x == NO and flip_y == YES:
        anchor.offset = [anchor.offset[0] - 0.1, anchor.offset[1] + 0.2]
    elif flip_x == YES and flip_y == YES:
        anchor.offset = [anchor.offset[0] - 0.1, anchor.offset[1] - 0.1]

    assert ij_to_s(anchor.offset, resolution, orientation) == s 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import json
from pathlib import Path
from a5.traversal.quintant_neighbors import get_cell_neighbors


def load_fixtures():
    fixture_path = Path(__file__).parent / "fixtures" / "quintant-neighbors.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


class TestGetCellNeighbors:
    def test_quintant_neighbors_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures:
            inp = case["input"]
            expected = case["output"]["neighbors"]
            result = get_cell_neighbors(inp["s"], inp["resolution"], inp["orientation"])
            assert result == expected, \
                f's={inp["s"]}, res={inp["resolution"]}, orient={inp["orientation"]}: {result} != {expected}'



"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

from typing import NewType, Tuple, TypeVar, Union

# Base types
Degrees = NewType('Degrees', float)
Radians = NewType('Radians', float)

# Vector types
Vec2 = Tuple[float, float]  # 2D vector
Vec3 = Tuple[float, float, float]  # 3D vector

# 2D coordinate systems
"""
2D cartesian coordinate system with origin at the center of a dodecahedron face
"""
Face = NewType('Face', Vec2)

"""
2D polar coordinate system with origin at the center of a dodecahedron face
"""
Polar = NewType('Polar', Tuple[float, Radians])

"""
2D planar coordinate system defined by the eigenvectors of the lattice tiling
"""
IJ = NewType('IJ', Vec2)

"""
2D planar coordinate system formed by the transformation K -> I + J
"""
KJ = NewType('KJ', Vec2)

# 3D coordinate systems
"""
3D cartesian system centered on unit sphere/dodecahedron
"""
Cartesian = NewType('Cartesian', Vec3)

"""
3D spherical coordinate system centered on unit sphere/dodecahedron
"""
Spherical = NewType('Spherical', Tuple[Radians, Radians])

"""
Geographic longitude & latitude
"""
LonLat = NewType('LonLat', Tuple[Degrees, Degrees])

# Barycentric coordinates
"""
Barycentric coordinates for a triangle (sum to 1)
"""
Barycentric = NewType('Barycentric', Tuple[float, float, float])

"""
Triangle defined by three face coordinates
"""
FaceTriangle = NewType('FaceTriangle', Tuple[Face, Face, Face])

"""
Triangle defined by three spherical coordinates
"""
SphericalTriangle = NewType('SphericalTriangle', Tuple[Cartesian, Cartesian, Cartesian]) 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import pytest
import json
import math
from pathlib import Path
from a5.geometry.spherical_triangle import SphericalTriangleShape
from a5.core.coordinate_systems import Cartesian
from a5.math.vec3 import length
from tests.matchers import is_close_array, is_close

# Load test fixtures
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "spherical-triangle.json") as f:
    FIXTURES = json.load(f)


class TestSphericalTriangleConstructor:
    """Test constructor functionality"""
    
    def test_requires_exactly_3_vertices(self):
        """Test that constructor requires exactly 3 vertices"""
        with pytest.raises(ValueError, match="SphericalTriangleShape requires exactly 3 vertices"):
            SphericalTriangleShape([])
        
        with pytest.raises(ValueError, match="SphericalTriangleShape requires exactly 3 vertices"):
            SphericalTriangleShape([
                (1, 0, 0), 
                (0, 1, 0)
            ])
        
        with pytest.raises(ValueError, match="SphericalTriangleShape requires exactly 3 vertices"):
            SphericalTriangleShape([
                (1, 0, 0), 
                (0, 1, 0), 
                (0, 0, 1), 
                (1, 1, 1)
            ])

    def test_accepts_exactly_3_vertices(self):
        """Test that constructor accepts exactly 3 vertices"""
        # Should not raise an exception
        triangle = SphericalTriangleShape([
            (1, 0, 0), 
            (0, 1, 0), 
            (0, 0, 1)
        ])
        assert len(triangle.vertices) == 3


class TestSphericalTriangleGetBoundary:
    """Test getBoundary functionality"""
    
    def test_returns_boundary_points_with_different_segment_counts(self):
        """Test boundary points with different segment counts"""
        for i, fixture in enumerate(FIXTURES):
            triangle = SphericalTriangleShape([
                tuple(vertex) for vertex in fixture["vertices"]
            ])
            
            # Test boundaries with 1-3 segments
            for n_segments in [1, 2, 3]:
                boundary = triangle.get_boundary(n_segments, True)
                expected_boundary = fixture[f"boundary{n_segments}"]
                
                assert len(boundary) == len(expected_boundary), \
                    f"Fixture {i}, segments {n_segments}: expected {len(expected_boundary)} points, got {len(boundary)}"
                
                for j, point in enumerate(boundary):
                    assert is_close_array(list(point), expected_boundary[j], 6), \
                        f"Fixture {i}, segments {n_segments}, point {j}: expected {expected_boundary[j]}, got {list(point)}"


class TestSphericalTriangleSlerp:
    """Test slerp functionality"""
    
    def test_interpolates_between_vertices(self):
        """Test spherical linear interpolation between vertices"""
        for i, fixture in enumerate(FIXTURES):
            triangle = SphericalTriangleShape([
                tuple(vertex) for vertex in fixture["vertices"]
            ])
            
            for test_case in fixture["slerpTests"]:
                t = test_case["t"]
                expected = test_case["result"]
                
                actual = triangle.slerp(t)
                assert is_close_array(list(actual), expected, 6), \
                    f"Fixture {i}, t={t}: expected {expected}, got {list(actual)}"
                
                # Should be normalized
                vector_length = length(actual)
                assert abs(vector_length - 1) < 1e-10, \
                    f"Fixture {i}, t={t}: result not normalized, length={vector_length}"


class TestSphericalTriangleContainsPoint:
    """Test containsPoint functionality"""
    
    def test_correctly_identifies_points_inside_and_outside_triangle(self):
        """Test point containment detection"""
        for i, fixture in enumerate(FIXTURES):
            triangle = SphericalTriangleShape([
                tuple(vertex) for vertex in fixture["vertices"]
            ])
            
            for test_case in fixture["containsPointTests"]:
                point = tuple(test_case["point"])
                expected = test_case["result"]
                
                actual = triangle.contains_point(point)
                assert is_close(actual, expected, 6), \
                    f"Fixture {i}, point={list(point)}: expected {expected}, got {actual}"


class TestSphericalTriangleGetArea:
    """Test getArea functionality"""
    
    def test_returns_correct_area_for_all_triangles(self):
        """Test area calculation for all triangles"""
        for i, fixture in enumerate(FIXTURES):
            triangle = SphericalTriangleShape([
                tuple(vertex) for vertex in fixture["vertices"]
            ])
            
            area = triangle.get_area()
            expected_area = fixture["area"]
            
            assert is_close(area, expected_area, 6), \
                f"Fixture {i}: expected area {expected_area}, got {area}"
            
            # Area can be negative for some winding orders, so check absolute value
            assert abs(area) > 0, \
                f"Fixture {i}: area should be non-zero, got {area}"
            assert abs(area) <= 2 * math.pi, \
                f"Fixture {i}: area should be <= 2π, got {abs(area)}" 


"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import cast
from ..core.coordinate_systems import Polar, Spherical

class GnomonicProjection:
    """
    Gnomonic projection implementation that converts between spherical and polar coordinates.
    """
    
    def forward(self, spherical: Spherical) -> Polar:
        """
        Projects spherical coordinates to polar coordinates using gnomonic projection
        
        Args:
            spherical: Spherical coordinates [theta, phi]
            
        Returns:
            Polar coordinates [rho, gamma]
        """
        theta, phi = spherical
        return cast(Polar, (math.tan(phi), theta))

    def inverse(self, polar: Polar) -> Spherical:
        """
        Unprojects polar coordinates to spherical coordinates using gnomonic projection
        
        Args:
            polar: Polar coordinates [rho, gamma]
            
        Returns:
            Spherical coordinates [theta, phi]
        """
        rho, gamma = polar
        return cast(Spherical, (gamma, math.atan(rho)))

__all__ = ['GnomonicProjection'] 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from .pentagon import PentagonShape
from .spherical_polygon import SphericalPolygonShape, SphericalPolygon

__all__ = ['PentagonShape', 'SphericalPolygonShape', 'SphericalPolygon'] 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from typing import List
from ..core.coordinate_systems import Cartesian
from .spherical_polygon import SphericalPolygonShape

# Type alias for clarity
SphericalTriangle = List[Cartesian]


class SphericalTriangleShape(SphericalPolygonShape):
    """
    A spherical triangle is a spherical polygon with exactly 3 vertices
    """
    
    def __init__(self, vertices: SphericalTriangle):
        if len(vertices) != 3:
            raise ValueError('SphericalTriangleShape requires exactly 3 vertices')
        super().__init__(vertices) 


"""
Tests for hex conversion utilities.
"""

import pytest
from a5.core.hex import hex_to_u64, u64_to_hex


def test_hex_to_u64():
    """Test hex to u64 conversion."""
    assert hex_to_u64("1a2b3c") == 1715004
    assert hex_to_u64("0") == 0
    assert hex_to_u64("ff") == 255
    assert hex_to_u64("ffffffff") == 4294967295


def test_u64_to_hex():
    """Test u64 to hex conversion."""
    assert u64_to_hex(1715004) == "1a2b3c"
    assert u64_to_hex(0) == "0"
    assert u64_to_hex(255) == "ff"
    assert u64_to_hex(4294967295) == "ffffffff"


def test_round_trip():
    """Test that converting back and forth gives the same result."""
    test_values = ["1a2b3c", "0", "ff", "ffffffff"]
    for hex_str in test_values:
        u64_value = hex_to_u64(hex_str)
        result = u64_to_hex(u64_value)
        assert result == hex_str 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

"""
Optimized implementation of compact/uncompact functions for A5 DGGS.

This version uses cell_to_children for expansion and stride-based sibling detection
for compaction.
"""

from typing import List

from .serialization import (
    get_resolution,
    cell_to_children,
    cell_to_parent,
    get_stride,
    is_first_child,
    FIRST_HILBERT_RESOLUTION
)
from .cell_info import get_num_children


def uncompact(cells: List[int], target_resolution: int) -> List[int]:
    """
    Expands a set of A5 cells to a target resolution by generating all descendant cells.

    Args:
        cells: List of A5 cell identifiers to uncompact
        target_resolution: The target resolution level for all output cells

    Returns:
        List of cell identifiers, all at the target resolution
    """
    # First calculate how much space is needed
    n = 0
    resolutions = []
    for cell in cells:
        resolution = get_resolution(cell)
        resolution_diff = target_resolution - resolution
        if resolution_diff < 0:
            raise ValueError(
                f"Cannot uncompact cell at resolution {resolution} to lower resolution {target_resolution}"
            )

        resolutions.append(resolution)
        n += get_num_children(resolution, target_resolution)

    # Write directly into pre-allocated list
    result = [0] * n
    offset = 0
    for i, cell in enumerate(cells):
        resolution = resolutions[i]

        num_children = get_num_children(resolution, target_resolution)
        if num_children == 1:
            result[offset] = cell
        else:
            children = cell_to_children(cell, target_resolution)
            for j, child in enumerate(children):
                result[offset + j] = child

        offset += num_children

    return result


def compact(cells: List[int]) -> List[int]:
    """
    Compacts a set of A5 cells by replacing complete groups of sibling cells with their parent cells.

    Args:
        cells: List of A5 cell identifiers to compact

    Returns:
        List of compacted cell identifiers (typically smaller than input)
    """
    if len(cells) == 0:
        return []

    # Single sort and dedup
    current_cells = sorted(set(cells))

    # Compact until no more changes
    # No re-sorting needed - parents maintain sorted order!
    changed = True
    while changed:
        changed = False
        result = []
        i = 0

        while i < len(current_cells):
            cell = current_cells[i]
            resolution = get_resolution(cell)

            # Can't compact below resolution 0
            if resolution < 0:
                result.append(cell)
                i += 1
                continue

            # Check for complete sibling group using unified stride-based approach
            if resolution >= FIRST_HILBERT_RESOLUTION:
                expected_children = 4  # Hilbert levels have 4 siblings
            elif resolution == 0:
                expected_children = 12  # First two levels are exceptions, with 12 & 5 siblings
            else:
                expected_children = 5

            if i + expected_children <= len(current_cells):
                has_all_siblings = True

                # Use stride-based checking for all resolutions
                # First check if this cell is a first child (at a sibling group boundary)
                if is_first_child(cell, resolution):
                    stride = get_stride(resolution)

                    # Check that all expected siblings are present with correct stride
                    for j in range(1, expected_children):
                        expected_cell = cell + j * stride
                        if current_cells[i + j] != expected_cell:
                            has_all_siblings = False
                            break
                else:
                    # First cell is not at a sibling group boundary
                    has_all_siblings = False

                if has_all_siblings:
                    # Compute parent only once when needed
                    parent = cell_to_parent(cell)
                    result.append(parent)
                    i += expected_children
                    changed = True
                    continue

            result.append(cell)
            i += 1

        current_cells = result

    return current_cells



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import pytest
import json
from pathlib import Path
from a5.projections.gnomonic import GnomonicProjection
from a5.core.coordinate_systems import Polar, Spherical
from tests.matchers import is_close_array

# Load test fixtures
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "gnomonic.json") as f:
    TEST_DATA = json.load(f)


@pytest.fixture
def gnomonic():
    return GnomonicProjection()

def test_forward_projections(gnomonic):
    """Test forward projections match expected values"""
    for test_case in TEST_DATA["forward"]:
        result = gnomonic.forward(tuple(test_case["input"]))
        assert is_close_array(result, test_case["expected"]), \
            f"Expected {test_case['expected']}, got {result}"

def test_round_trip_forward_projections(gnomonic):
    """Test forward projections can be reversed accurately"""
    for test_case in TEST_DATA["forward"]:
        spherical = tuple(test_case["input"])
        polar = gnomonic.forward(spherical)
        result = gnomonic.inverse(polar)
        assert is_close_array(result, spherical), \
            f"Expected {spherical}, got {result}"

def test_inverse_projections(gnomonic):
    """Test inverse projections match expected values"""
    for test_case in TEST_DATA["inverse"]:
        result = gnomonic.inverse(tuple(test_case["input"]))
        assert is_close_array(result, test_case["expected"]), \
            f"Expected {test_case['expected']}, got {result}"

def test_round_trip_inverse_projections(gnomonic):
    """Test inverse projections can be reversed accurately"""
    for test_case in TEST_DATA["inverse"]:
        polar = tuple(test_case["input"])
        spherical = gnomonic.inverse(polar)
        result = gnomonic.forward(spherical)
        assert is_close_array(result, polar), \
            f"Expected {polar}, got {result}" 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import pytest
import json
import math
from pathlib import Path
from a5.projections.polyhedral import PolyhedralProjection
from a5.core.coordinate_systems import Cartesian
from a5.math.vec3 import length
from tests.matchers import is_close_array

# Load test fixtures
FIXTURES_DIR = Path(__file__).parent / "fixtures"
with open(FIXTURES_DIR / "polyhedral.json") as f:
    TEST_DATA = json.load(f)

# Extract static data from test data
TEST_SPHERICAL_TRIANGLE = TEST_DATA["static"]["TEST_SPHERICAL_TRIANGLE"]
TEST_FACE_TRIANGLE = TEST_DATA["static"]["TEST_FACE_TRIANGLE"]

AUTHALIC_RADIUS = 6371.0072  # km
def dot_product(a, b):
    return sum(x * y for x, y in zip(a, b))

MAX_ANGLE = max(
    math.acos(max(-1, min(1, dot_product(TEST_SPHERICAL_TRIANGLE[0], TEST_SPHERICAL_TRIANGLE[1])))),
    math.acos(max(-1, min(1, dot_product(TEST_SPHERICAL_TRIANGLE[1], TEST_SPHERICAL_TRIANGLE[2])))),
    math.acos(max(-1, min(1, dot_product(TEST_SPHERICAL_TRIANGLE[2], TEST_SPHERICAL_TRIANGLE[0]))))
)
MAX_ARC_LENGTH_MM = AUTHALIC_RADIUS * MAX_ANGLE * 1e9
DESIRED_MM_PRECISION = 0.01


@pytest.fixture
def polyhedral():
    return PolyhedralProjection()

class TestPolyhedralProjectionForward:
    """Test forward projection functionality"""
    
    def test_forward_projections(self, polyhedral):
        """Test forward projections match expected values"""
        for test_case in TEST_DATA["forward"]:
            result = polyhedral.forward(
                test_case["input"],
                TEST_SPHERICAL_TRIANGLE, 
                TEST_FACE_TRIANGLE
            )
            assert is_close_array(list(result), test_case["expected"]), \
                f"Expected {test_case['expected']}, got {list(result)}"

    def test_round_trip_forward_projections(self, polyhedral):
        """Test round trip forward projections"""
        largest_error = 0
        
        for test_case in TEST_DATA["forward"]:
            spherical = test_case["input"]
            polar = polyhedral.forward(spherical, TEST_SPHERICAL_TRIANGLE, TEST_FACE_TRIANGLE)
            result = polyhedral.inverse(polar, TEST_FACE_TRIANGLE, TEST_SPHERICAL_TRIANGLE)
            error = length([r - s for r, s in zip(result, spherical)])
            largest_error = max(largest_error, error)
            assert is_close_array(list(result), spherical), \
                f"Round trip failed: expected {spherical}, got {list(result)}"
        
        # Check precision requirement
        assert largest_error * MAX_ARC_LENGTH_MM < DESIRED_MM_PRECISION, \
            f"Accuracy requirement not met: {largest_error * MAX_ARC_LENGTH_MM:.6f}mm > {DESIRED_MM_PRECISION}mm"


class TestPolyhedralProjectionInverse:
    """Test inverse projection functionality"""
    
    def test_inverse_projections(self, polyhedral):
        """Test inverse projections match expected values"""
        for test_case in TEST_DATA["inverse"]:
            result = polyhedral.inverse(
                test_case["input"],
                TEST_FACE_TRIANGLE,
                TEST_SPHERICAL_TRIANGLE
            )
            assert is_close_array(list(result), test_case["expected"]), \
                f"Expected {test_case['expected']}, got {list(result)}"

    def test_round_trip_inverse_projections(self, polyhedral):
        """Test round trip inverse projections"""
        for test_case in TEST_DATA["inverse"]:
            face_point = test_case["input"]
            spherical = polyhedral.inverse(face_point, TEST_FACE_TRIANGLE, TEST_SPHERICAL_TRIANGLE)
            result = polyhedral.forward(spherical, TEST_SPHERICAL_TRIANGLE, TEST_FACE_TRIANGLE)
            assert is_close_array(list(result), face_point), \
                f"Round trip failed: expected {face_point}, got {list(result)}" 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

"""
Quaternion operations following gl-matrix API patterns
https://glmatrix.net/docs/module-quat.html

Portions of this code are derived from https://github.com/toji/gl-matrix
"""

import math
from typing import List, Union
from . import vec3

# Type alias for quaternions - [x, y, z, w]
Quat = Union[List[float], tuple]

def create() -> List[float]:
    """
    Creates a new identity quat
    
    Returns:
        A new quaternion [x, y, z, w]
    """
    return [0.0, 0.0, 0.0, 1.0]

def length(a: Quat) -> float:
    """
    Calculates the length of a quat
    
    Args:
        a: quaternion to calculate length of
        
    Returns:
        length of a
    """
    x, y, z, w = a[0], a[1], a[2], a[3]
    return math.sqrt(x * x + y * y + z * z + w * w)

# Alias for length
len = length

def conjugate(out: Quat, a: Quat) -> Quat:
    """
    Calculates the conjugate of a quat
    If the quaternion is normalized, this function is faster than quat.inverse and produces the same result.
    
    Args:
        out: the receiving quaternion
        a: quat to calculate conjugate of
        
    Returns:
        out
    """
    out[0] = -a[0]
    out[1] = -a[1]
    out[2] = -a[2]
    out[3] = a[3]
    return out




"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import cast, Literal, TypedDict, Dict
from .coordinate_systems import Radians

# Golden ratio
PHI = (1 + math.sqrt(5)) / 2

TWO_PI = cast(Radians, 2 * math.pi)
TWO_PI_OVER_5 = cast(Radians, 2 * math.pi / 5)
PI_OVER_5 = cast(Radians, math.pi / 5)
PI_OVER_10 = cast(Radians, math.pi / 10)

# Angles between faces
dihedral_angle = cast(Radians, 2 * math.atan(PHI))  # Angle between pentagon faces (radians) = 116.565°
interhedral_angle = cast(Radians, math.pi - dihedral_angle)  # Angle between pentagon faces (radians) = 63.435°
face_edge_angle = cast(Radians, -0.5 * math.pi + math.acos(-1 / math.sqrt(3 - PHI)))  # = 58.28252558853899

# Distance from center to edge of pentagon face
distance_to_edge = (math.sqrt(5) - 1) / 2  # PHI - 1
distance_to_vertex = 3 - math.sqrt(5)  # 2 * (2 - PHI)

"""
Radius of the inscribed sphere in dodecahedron
"""
R_INSCRIBED = 1.0

"""
Radius of the sphere that touches the dodecahedron's edge midpoints
"""
R_MIDEDGE = math.sqrt(3 - PHI)

"""
Radius of the circumscribed sphere for dodecahedron
"""
R_CIRCUMSCRIBED = math.sqrt(3) * R_MIDEDGE / PHI

# Authalic radius of Earth (meters)
AUTHALIC_RADIUS_EARTH = 6371007.2  # m

# Authalic surface area of Earth (square meters)
AUTHALIC_AREA_EARTH = 4 * math.pi * AUTHALIC_RADIUS_EARTH * AUTHALIC_RADIUS_EARTH  # m^2

__all__ = [
    'PHI',
    'TWO_PI',
    'TWO_PI_OVER_5',
    'PI_OVER_5',
    'PI_OVER_10',
    'dihedral_angle',
    'interhedral_angle',
    'face_edge_angle',
    'distance_to_edge',
    'distance_to_vertex',
    'R_INSCRIBED',
    'R_MIDEDGE',
    'R_CIRCUMSCRIBED',
    'AUTHALIC_RADIUS_EARTH',
    'AUTHALIC_AREA_EARTH'
] 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from typing import List, Tuple
from .types import Quaternary, Flip


def reverse_pattern(pattern: List[int]) -> List[int]:
    """Reverse a pattern by finding the index of each position."""
    return [pattern.index(i) for i in range(len(pattern))]


# Patterns used to rearrange the cells when shifting
PATTERN = [0, 1, 3, 4, 5, 6, 7, 2]
PATTERN_FLIPPED = [0, 1, 2, 7, 3, 4, 5, 6]
PATTERN_REVERSED = reverse_pattern(PATTERN)
PATTERN_FLIPPED_REVERSED = reverse_pattern(PATTERN_FLIPPED)


def shift_digits(digits: List[int], i: int, flips: List[int], invert_j: bool, pattern: List[int]) -> None:
    """Shift digits based on pattern to adjust cell layout."""
    if i <= 0:
        return

    parent_k = digits[i] if i < len(digits) else 0
    child_k = digits[i - 1]
    F = flips[0] + flips[1]

    # Detect when cells need to be shifted
    needs_shift = True
    first = True

    # The value of F which cells need to be shifted
    # The rule is flipped depending on the orientation, specifically on the value of invert_j
    if invert_j != (F == 0):
        needs_shift = parent_k in (1, 2)  # Second & third pentagons only
        first = parent_k == 1  # Second pentagon is first
    else:
        needs_shift = parent_k < 2  # First two pentagons only
        first = parent_k == 0  # First pentagon is first

    if not needs_shift:
        return

    # Apply the pattern by setting the digits based on the value provided
    src = child_k if first else child_k + 4
    dst = pattern[src]
    digits[i - 1] = dst % 4
    digits[i] = (parent_k + 4 + (dst // 4) - (src // 4)) % 4



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import json
from pathlib import Path
from a5.traversal.global_neighbors import get_global_cell_neighbors


def load_fixtures():
    fixture_path = Path(__file__).parent / "fixtures" / "global-neighbors.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


def hex_to_int(hex_str: str) -> int:
    return int(hex_str, 16)


class TestGetGlobalCellNeighbors:
    def test_global_neighbors_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures:
            cell_id = hex_to_int(case["input"]["cellId"])
            expected = sorted(hex_to_int(h) for h in case["output"]["neighbors"])
            result = get_global_cell_neighbors(cell_id)
            assert result == expected, \
                f'cellId={case["input"]["cellId"]}: got {[hex(c) for c in result]}, expected {[hex(c) for c in expected]}'

    def test_global_edge_neighbors_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures:
            cell_id = hex_to_int(case["input"]["cellId"])
            expected = sorted(hex_to_int(h) for h in case["output"]["edgeNeighbors"])
            result = get_global_cell_neighbors(cell_id, edge_only=True)
            assert result == expected, \
                f'cellId={case["input"]["cellId"]}: got {[hex(c) for c in result]}, expected {[hex(c) for c in expected]}'



import pytest
import json
from pathlib import Path

from a5.geometry.pentagon import PentagonShape


def load_fixtures():
    """Load pentagon test fixtures."""
    fixture_path = Path(__file__).parent / "fixtures" / "pentagon.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


class TestPentagonShape:
    """Test cases for PentagonShape class."""
    
    def test_contains_point(self):
        """Test that containsPoint returns correct results for all test cases."""
        fixtures = load_fixtures()
        
        for fixture in fixtures:
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            
            for test_case in fixture["containsPointTests"]:
                point = tuple(test_case["point"])
                expected = test_case["result"]
                actual = pentagon.contains_point(point)
                assert abs(actual - expected) < 1e-6, f"Contains point test failed: expected {expected}, got {actual}"

    def test_get_area(self):
        """Test that getArea returns correct area for all pentagons."""
        fixtures = load_fixtures()
        
        for fixture in fixtures:
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            area = pentagon.get_area()
            expected = fixture["area"]
            assert abs(area - expected) < 1e-6, f"Area test failed: expected {expected}, got {area}"

    def test_get_center(self):
        """Test that getCenter returns correct center for all pentagons."""
        fixtures = load_fixtures()
        
        for fixture in fixtures:
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            center = pentagon.get_center()
            expected = fixture["center"]
            assert abs(center[0] - expected[0]) < 1e-6, f"Center X test failed: expected {expected[0]}, got {center[0]}"
            assert abs(center[1] - expected[1]) < 1e-6, f"Center Y test failed: expected {expected[1]}, got {center[1]}"

    def test_transformations(self):
        """Test that all transformations work correctly."""
        fixtures = load_fixtures()
        
        for fixture in fixtures:
            # Test scale transformation
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            scaled = pentagon.clone().scale(2)
            vertices = scaled.get_vertices()
            
            expected_scale = fixture["transformTests"]["scale"]
            for i, expected in enumerate(expected_scale):
                assert abs(vertices[i][0] - expected[0]) < 1e-6, f"Scale X test failed at vertex {i}"
                assert abs(vertices[i][1] - expected[1]) < 1e-6, f"Scale Y test failed at vertex {i}"

            # Test rotate180 transformation
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            rotated = pentagon.clone().rotate180()
            vertices = rotated.get_vertices()
            
            expected_rotate = fixture["transformTests"]["rotate180"]
            for i, expected in enumerate(expected_rotate):
                assert abs(vertices[i][0] - expected[0]) < 1e-6, f"Rotate180 X test failed at vertex {i}"
                assert abs(vertices[i][1] - expected[1]) < 1e-6, f"Rotate180 Y test failed at vertex {i}"

            # Test reflectY transformation
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            reflected = pentagon.clone().reflect_y()
            vertices = reflected.get_vertices()
            
            expected_reflect = fixture["transformTests"]["reflectY"]
            for i, expected in enumerate(expected_reflect):
                assert abs(vertices[i][0] - expected[0]) < 1e-6, f"ReflectY X test failed at vertex {i}"
                assert abs(vertices[i][1] - expected[1]) < 1e-6, f"ReflectY Y test failed at vertex {i}"

            # Test translate transformation
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            translated = pentagon.clone().translate((1, 1))
            vertices = translated.get_vertices()
            
            expected_translate = fixture["transformTests"]["translate"]
            for i, expected in enumerate(expected_translate):
                assert abs(vertices[i][0] - expected[0]) < 1e-6, f"Translate X test failed at vertex {i}"
                assert abs(vertices[i][1] - expected[1]) < 1e-6, f"Translate Y test failed at vertex {i}"

    def test_split_edges(self):
        """Test that splitEdges returns split edges with different segment counts."""
        fixtures = load_fixtures()
        
        for fixture in fixtures:
            pentagon = PentagonShape([tuple(vertex) for vertex in fixture["vertices"]])
            
            # Test boundaries with 2-3 segments
            for n_segments in [2, 3]:
                split = pentagon.clone().split_edges(n_segments)
                vertices = split.get_vertices()
                expected_vertices = fixture["splitEdgesTests"][f"segments{n_segments}"]
                
                assert len(vertices) == len(expected_vertices), f"Split edges vertex count mismatch for {n_segments} segments"
                
                for i, expected in enumerate(expected_vertices):
                    assert abs(vertices[i][0] - expected[0]) < 1e-6, f"Split edges X test failed at vertex {i} for {n_segments} segments"
                    assert abs(vertices[i][1] - expected[1]) < 1e-6, f"Split edges Y test failed at vertex {i} for {n_segments} segments" 


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from typing import Tuple, List, Literal

Orientation = Literal['uv', 'vu', 'uw', 'wu', 'vw', 'wv']

Quaternary = Literal[0, 1, 2, 3]
YES: Literal[-1] = -1
NO: Literal[1] = 1
Flip = Literal[-1, 1]


class Anchor:
    def __init__(self, q: Quaternary, offset: Tuple[float, float], flips: Tuple[Flip, Flip]):
        self.q = q
        self.offset = offset
        self.flips = flips



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import math
from .constants import AUTHALIC_AREA_EARTH

AUTHALIC_AREA = AUTHALIC_AREA_EARTH
FIRST_HILBERT_RESOLUTION = 2

def get_num_cells(resolution: int) -> int:
    """
    Returns the number of cells at a given resolution.

    Args:
        resolution: The resolution level

    Returns:
        Number of cells at the given resolution
    """
    if resolution < 0:
        return 0
    if resolution == 0:
        return 12
    return 60 * (4 ** (resolution - 1))


def get_num_children(parent_resolution: int, child_resolution: int) -> int:
    """
    Returns the number of children between two resolutions.

    Args:
        parent_resolution: The parent resolution level
        child_resolution: The child resolution level

    Returns:
        Number of children
    """
    if child_resolution < parent_resolution:
        return 0
    if child_resolution == parent_resolution:
        return 1
    if parent_resolution >= FIRST_HILBERT_RESOLUTION:
        # Between levels of constant aperture of 4, relation simplifies
        return 4 ** (child_resolution - parent_resolution)

    parent_count = get_num_cells(parent_resolution) or 1
    child_count = get_num_cells(child_resolution)
    return child_count // parent_count


def cell_area(resolution: int) -> float:
    """
    Returns the area of a cell at a given resolution in square meters.

    Args:
        resolution: The resolution level

    Returns:
        Area of a cell in square meters
    """
    if resolution < 0:
        return AUTHALIC_AREA
    return AUTHALIC_AREA / get_num_cells(resolution)



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import json
from pathlib import Path
from a5.traversal.cap import meters_to_h, estimate_cell_radius, pick_coarse_resolution, spherical_cap
from a5.core.serialization import get_resolution
from a5.core.compact import uncompact


def load_fixtures():
    fixture_path = Path(__file__).parent / "fixtures" / "cap.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


def hex_to_int(hex_str: str) -> int:
    return int(hex_str, 16)


class TestMetersToH:
    def test_meters_to_h_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["helpers"]["metersToH"]:
            result = meters_to_h(case["meters"])
            # Allow 1 ULP tolerance: Python's math.sin may round differently from JS Math.sin
            assert abs(result - case["expectedH"]) < 1e-15, \
                f'meters={case["meters"]}: {result} != {case["expectedH"]}'


class TestEstimateCellRadius:
    def test_estimate_cell_radius_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["helpers"]["estimateCellRadius"]:
            result = estimate_cell_radius(case["resolution"])
            # Allow tiny tolerance: Python's math may round differently from JS
            assert abs(result - case["expectedMeters"]) < 1e-6, \
                f'res={case["resolution"]}: {result} != {case["expectedMeters"]}'

    def test_decreases_with_resolution(self):
        fixtures = load_fixtures()
        cases = fixtures["helpers"]["estimateCellRadius"]
        for i in range(1, len(cases)):
            assert cases[i]["expectedMeters"] < cases[i - 1]["expectedMeters"]


class TestPickCoarseResolution:
    def test_pick_coarse_resolution_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["helpers"]["pickCoarseResolution"]:
            result = pick_coarse_resolution(case["radius"], case["targetRes"])
            assert result == case["expectedCoarseRes"], \
                f'radius={case["radius"]}, targetRes={case["targetRes"]}: {result} != {case["expectedCoarseRes"]}'

    def test_never_exceeds_target_res(self):
        fixtures = load_fixtures()
        for case in fixtures["helpers"]["pickCoarseResolution"]:
            result = pick_coarse_resolution(case["radius"], case["targetRes"])
            assert result <= case["targetRes"]


class TestSphericalCap:
    def test_spherical_cap_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["sphericalCap"]:
            cell_id = hex_to_int(case["cellId"])
            radius = case["radius"]
            target_res = get_resolution(cell_id)
            expected = sorted(hex_to_int(h) for h in case["cells"])
            result = sorted(uncompact(spherical_cap(cell_id, radius), target_res))
            assert result == expected, \
                f'cellId={case["cellId"]}, radius={radius}: got {len(result)} cells, expected {len(expected)}'

    def test_spherical_cap_compact_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["sphericalCapCompact"]:
            cell_id = hex_to_int(case["cellId"])
            radius = case["radius"]
            expected = sorted(hex_to_int(h) for h in case["compactedCells"])
            result = sorted(spherical_cap(cell_id, radius))
            assert result == expected, \
                f'cellId={case["cellId"]}, radius={radius}: got {len(result)} cells, expected {len(expected)}'



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from typing import List, Tuple, Optional, Set
from ..lattice import (
    Orientation, Triple,
    s_to_anchor, anchor_to_triple, triple_to_anchor, triple_to_s, triple_parity, triple_in_bounds
)
from ..core.utils import Origin
from ..core.serialization import deserialize, serialize, FIRST_HILBERT_RESOLUTION
from ..core.origin import segment_to_quintant, quintant_to_segment, origins
from ..core.face_adjacency import FACE_ADJACENCY
from .quintant_neighbors import find_quintant_neighbor_s

# Neighbor delta: (dx, dy, dz, is_edge_sharing)
NeighborDelta = Tuple[int, int, int, bool]

# Cross-quintant left-edge deltas (source z=0), indexed by parity * 2 + (y_odd ? 1 : 0)
LEFT_EDGE_DELTAS: List[List[NeighborDelta]] = [
    # parity=0, yEven
    [(0, 0, 0, True), (0, 0, 1, False)],
    # parity=0, yOdd
    [(0, 0, 0, True), (0, 1, 0, True), (0, -1, 1, False), (0, 1, -1, False)],
    # parity=1, yEven
    [],
    # parity=1, yOdd
    [(0, -1, 0, True), (0, 0, -1, False)],
]

# Cross-quintant right-edge deltas (source x=0), indexed by parity * 2 + (y_odd ? 1 : 0)
RIGHT_EDGE_DELTAS: List[List[NeighborDelta]] = [
    # parity=0, yEven
    [(0, 0, 0, True), (0, 1, 0, True), (-1, 1, 0, False), (1, -1, 0, False)],
    # parity=0, yOdd
    [(0, 0, 0, True), (1, 0, 0, False)],
    # parity=1, yEven
    [(0, -1, 0, True), (-1, 0, 0, False)],
    # parity=1, yOdd
    [],
]

# Cross-face base-edge deltas (source y=maxRow), indexed by parity
CROSS_FACE_DELTAS: List[List[NeighborDelta]] = [
    # parity=0
    [(0, 0, 0, True), (1, 0, 0, True), (1, 0, -1, False)],
    # parity=1
    [(0, 0, -1, True), (0, 0, 0, False)],
]


class _NeighborContext:
    """Shared state for neighbor search helpers."""
    def __init__(self, hilbert_res: int, resolution: int, edge_only: bool):
        self.hilbert_res = hilbert_res
        self.resolution = resolution
        self.max_s = 4 ** hilbert_res
        self.max_row = (1 << hilbert_res) - 1
        self.edge_only = edge_only
        self.neighbor_set: Set[int] = set()


def _add_neighbor(
    ctx: _NeighborContext,
    neighbor_triple: Triple, orientation: Orientation,
    neighbor_origin: Origin, neighbor_segment: int
) -> None:
    """Try to convert a triple to a cell ID and add it to the neighbor set."""
    s = triple_to_s(neighbor_triple, ctx.hilbert_res, orientation)
    if s is None or s < 0 or s >= ctx.max_s:
        return
    ctx.neighbor_set.add(serialize({
        'origin': neighbor_origin, 'segment': neighbor_segment,
        'S': s, 'resolution': ctx.resolution
    }))


def _add_delta_neighbors(
    ctx: _NeighborContext,
    base: Triple, deltas: List[NeighborDelta],
    orientation: Orientation, neighbor_origin: Origin, neighbor_segment: int
) -> None:
    """Apply a delta table to a base triple and add valid neighbors."""
    for dx, dy, dz, is_edge in deltas:
        if ctx.edge_only and not is_edge:
            continue
        neighbor_triple = Triple(base.x + dx, base.y + dy, base.z + dz)
        if not triple_in_bounds(neighbor_triple, ctx.max_row):
            continue
        _add_neighbor(ctx, neighbor_triple, orientation, neighbor_origin, neighbor_segment)


def get_global_cell_neighbors(cell_id: int, edge_only: bool = False) -> List[int]:
    """
    Get all neighbors of a cell across quintant and face boundaries.

    Uses three strategies:
    - Within-quintant: Standard triple coordinate approach
    - Cross-quintant: Lateral edge handling via coordinate transformation
    - Cross-face: Base edge handling via x<->z swap
    """
    cell = deserialize(cell_id)
    origin, segment, S, resolution = cell['origin'], cell['segment'], cell['S'], cell['resolution']
    if resolution < FIRST_HILBERT_RESOLUTION:
        return []  # No neighbors for res 0-1

    hilbert_res = resolution - FIRST_HILBERT_RESOLUTION + 1
    source_quintant, source_orientation = segment_to_quintant(segment, origin)
    anchor = s_to_anchor(S, hilbert_res, source_orientation)

    # Triple coordinates are orientation-independent
    triple = anchor_to_triple(anchor)

    # Get uv anchor for is_neighbor validation (within-quintant)
    uv_source_anchor = triple_to_anchor(triple, hilbert_res, 'uv')

    ctx = _NeighborContext(hilbert_res, resolution, edge_only)

    # --- Within-quintant neighbors ---
    for neighbor_s in find_quintant_neighbor_s(triple, uv_source_anchor, S, hilbert_res, source_orientation, edge_only):
        ctx.neighbor_set.add(serialize({
            'origin': origin, 'segment': segment,
            'S': neighbor_s, 'resolution': resolution
        }))

    # --- Cross-quintant neighbors ---
    parity = triple_parity(triple)  # 0 or 1
    y_odd = triple.y % 2 != 0
    delta_index = parity * 2 + (1 if y_odd else 0)

    # Left edge (z=0): neighbor in previous quintant at swapped [0, y, x]
    if triple.z == 0:
        target_quintant = (source_quintant - 1 + 5) % 5
        target_segment, target_orientation = quintant_to_segment(target_quintant, origin)
        swapped_base = Triple(0, triple.y, triple.x)
        _add_delta_neighbors(ctx, swapped_base, LEFT_EDGE_DELTAS[delta_index], target_orientation, origin, target_segment)

    # Right edge (x=0): neighbor in next quintant at swapped [z, y, 0]
    if triple.x == 0:
        target_quintant = (source_quintant + 1) % 5
        target_segment, target_orientation = quintant_to_segment(target_quintant, origin)
        swapped_base = Triple(triple.z, triple.y, 0)
        _add_delta_neighbors(ctx, swapped_base, RIGHT_EDGE_DELTAS[delta_index], target_orientation, origin, target_segment)

    # --- Cross-face neighbors ---
    if triple.y == ctx.max_row:
        adj_face_id, adj_quintant = FACE_ADJACENCY[origin.id][source_quintant]
        adj_origin = origins[adj_face_id]
        adj_segment, adj_orientation = quintant_to_segment(adj_quintant, adj_origin)
        mirrored_base = Triple(triple.z, ctx.max_row, triple.x)
        _add_delta_neighbors(ctx, mirrored_base, CROSS_FACE_DELTAS[parity], adj_orientation, adj_origin, adj_segment)

    # Apex: [0,0,0] cells from all 5 quintants meet at the face center
    if triple.x == 0 and triple.y == 0 and triple.z == 0:
        for q in range(5):
            if q == source_quintant:
                continue
            # Adjacent quintants (distance=1) share an edge; non-adjacent (distance=2) share only a vertex
            distance = min(
                (q - source_quintant + 5) % 5,
                (source_quintant - q + 5) % 5
            )
            if edge_only and distance != 1:
                continue
            target_segment, target_orientation = quintant_to_segment(q, origin)
            _add_neighbor(ctx, triple, target_orientation, origin, target_segment)

    # Special case: base-left corner cells
    if triple.x == -ctx.max_row and triple.y == ctx.max_row and triple.z == 0:
        # Vertex neighbor 1: across the previous quintant's base edge
        prev_quintant = (source_quintant - 1 + 5) % 5
        prev_adj_face_id, prev_adj_quintant = FACE_ADJACENCY[origin.id][prev_quintant]
        prev_adj_origin = origins[prev_adj_face_id]
        prev_adj_segment, prev_adj_orientation = quintant_to_segment(prev_adj_quintant, prev_adj_origin)
        _add_neighbor(ctx, triple, prev_adj_orientation, prev_adj_origin, prev_adj_segment)

        # Vertex neighbor 2: adjacent quintant on the primary cross-face
        cross_face_id, cross_quintant = FACE_ADJACENCY[origin.id][source_quintant]
        cross_origin = origins[cross_face_id]
        next_cross_quintant = (cross_quintant + 1) % 5
        cross_segment, cross_orientation = quintant_to_segment(next_cross_quintant, cross_origin)
        _add_neighbor(ctx, triple, cross_orientation, cross_origin, cross_segment)

    result = sorted(ctx.neighbor_set)
    return result



"""
gl-matrix style vec3 operations for A5
Based on https://glmatrix.net/docs/module-vec3.html

All functions follow the gl-matrix convention of having an 'out' parameter
for the result, and return the 'out' parameter for chaining.
"""

import math
from typing import Tuple, Union, List, cast

# Type alias for 3D vectors - can be list or tuple
Vec3 = Union[List[float], Tuple[float, float, float]]

# Pre-allocated temporary vectors for performance (like TypeScript gl-matrix)
midpointAB = [0.0, 0.0, 0.0]
crossCD = [0.0, 0.0, 0.0]
scaledA = [0.0, 0.0, 0.0]
scaledB = [0.0, 0.0, 0.0]

def create() -> List[float]:
    """
    Creates a new vec3 initialized to [0, 0, 0]
    
    Returns:
        A new vec3
    """
    return [0.0, 0.0, 0.0]

def clone(a: Vec3) -> List[float]:
    """
    Creates a new vec3 initialized with values from an existing vector
    
    Args:
        a: vector to clone
        
    Returns:
        A new vec3
    """
    return [a[0], a[1], a[2]]

def copy(out: Vec3, a: Vec3) -> Vec3:
    """
    Copy the values from one vec3 to another
    
    Args:
        out: the receiving vector
        a: the source vector
        
    Returns:
        out
    """
    out[0] = a[0]
    out[1] = a[1]
    out[2] = a[2]
    return out

def set(out: Vec3, x: float, y: float, z: float) -> Vec3:
    """
    Set the components of a vec3 to the given values
    
    Args:
        out: the receiving vector
        x: X component
        y: Y component
        z: Z component
        
    Returns:
        out
    """
    out[0] = x
    out[1] = y
    out[2] = z
    return out

def add(out: Vec3, a: Vec3, b: Vec3) -> Vec3:
    """
    Adds two vec3's
    
    Args:
        out: the receiving vector
        a: the first operand
        b: the second operand
        
    Returns:
        out
    """
    out[0] = a[0] + b[0]
    out[1] = a[1] + b[1]
    out[2] = a[2] + b[2]
    return out

def subtract(out: Vec3, a: Vec3, b: Vec3) -> Vec3:
    """
    Subtracts vector b from vector a
    
    Args:
        out: the receiving vector
        a: the first operand
        b: the second operand
        
    Returns:
        out
    """
    out[0] = a[0] - b[0]
    out[1] = a[1] - b[1]
    out[2] = a[2] - b[2]
    return out

# Alias for TypeScript compatibility
sub = subtract

def scale(out: Vec3, a: Vec3, s: float) -> Vec3:
    """
    Scales a vec3 by a scalar number
    
    Args:
        out: the receiving vector
        a: the vector to scale
        s: amount to scale the vector by
        
    Returns:
        out
    """
    out[0] = a[0] * s
    out[1] = a[1] * s
    out[2] = a[2] * s
    return out

def dot(a: Vec3, b: Vec3) -> float:
    """
    Calculates the dot product of two vec3's
    
    Args:
        a: the first operand
        b: the second operand
        
    Returns:
        dot product of a and b
    """
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]

def cross(out: Vec3, a: Vec3, b: Vec3) -> Vec3:
    """
    Computes the cross product of two vec3's
    
    Args:
        out: the receiving vector
        a: the first operand
        b: the second operand
        
    Returns:
        out
    """
    ax, ay, az = a[0], a[1], a[2]
    bx, by, bz = b[0], b[1], b[2]
    
    out[0] = ay * bz - az * by
    out[1] = az * bx - ax * bz
    out[2] = ax * by - ay * bx
    return out

def length(a: Vec3) -> float:
    """
    Calculates the length of a vec3
    
    Args:
        a: vector to calculate length of
        
    Returns:
        length of a
    """
    x, y, z = a[0], a[1], a[2]
    return math.sqrt(x * x + y * y + z * z)

# Alias for length
len = length

def normalize(out: Vec3, a: Vec3) -> Vec3:
    """
    Normalize a vec3
    
    Args:
        out: the receiving vector
        a: vector to normalize
        
    Returns:
        out
    """
    x, y, z = a[0], a[1], a[2]
    len_sq = x * x + y * y + z * z
    
    if len_sq > 0:
        inv_len = 1.0 / math.sqrt(len_sq)
        out[0] = x * inv_len
        out[1] = y * inv_len
        out[2] = z * inv_len
    else:
        out[0] = 0
        out[1] = 0
        out[2] = 0
    
    return out

def distance(a: Vec3, b: Vec3) -> float:
    """
    Calculates the euclidean distance between two vec3's
    
    Args:
        a: the first operand
        b: the second operand
        
    Returns:
        distance between a and b
    """
    dx = b[0] - a[0]
    dy = b[1] - a[1]
    dz = b[2] - a[2]
    return math.sqrt(dx * dx + dy * dy + dz * dz)

def lerp(out: Vec3, a: Vec3, b: Vec3, t: float) -> Vec3:
    """
    Performs a linear interpolation between two vec3's
    
    Args:
        out: the receiving vector
        a: the first operand
        b: the second operand
        t: interpolation amount, in the range [0-1], between the two inputs
        
    Returns:
        out
    """
    ax, ay, az = a[0], a[1], a[2]
    out[0] = ax + t * (b[0] - ax)
    out[1] = ay + t * (b[1] - ay)
    out[2] = az + t * (b[2] - az)
    return out

def angle(a: Vec3, b: Vec3) -> float:
    """
    Get the angle between two 3D vectors
    
    Args:
        a: The first operand
        b: The second operand
        
    Returns:
        The angle in radians
    """
    # Normalize both vectors
    temp_a = normalize(create(), a)
    temp_b = normalize(create(), b)
    
    cos_angle = dot(temp_a, temp_b)
    
    # Clamp to avoid numerical errors
    cos_angle = max(-1.0, min(1.0, cos_angle))
    
    return math.acos(cos_angle)

def transformQuat(out: Vec3, a: Vec3, q: List[float]) -> Vec3:
    """
    Transforms the vec3 with a quat
    
    Args:
        out: the receiving vector
        a: the vector to transform
        q: quaternion to transform with [x, y, z, w]
        
    Returns:
        out
    """
    # Get quaternion components
    qx, qy, qz, qw = q[0], q[1], q[2], q[3]
    x, y, z = a[0], a[1], a[2]

    # Calculate cross product q × a
    uvx = qy * z - qz * y
    uvy = qz * x - qx * z
    uvz = qx * y - qy * x

    # Calculate cross product q × (q × a)
    uuvx = qy * uvz - qz * uvy
    uuvy = qz * uvx - qx * uvz
    uuvz = qx * uvy - qy * uvx

    # Scale uv by 2 * w
    w2 = qw * 2
    uvx *= w2
    uvy *= w2
    uvz *= w2

    # Scale uuv by 2
    uuvx *= 2
    uuvy *= 2
    uuvz *= 2

    # Add all components
    out[0] = x + uvx + uuvx
    out[1] = y + uvy + uuvy
    out[2] = z + uvz + uuvz
    return out

def tripleProduct(a: Vec3, b: Vec3, c: Vec3) -> float:
    """
    Computes the triple product of three vectors: a · (b × c)
    
    Args:
        a: first vector
        b: second vector  
        c: third vector
        
    Returns:
        scalar result a · (b × c)
    """
    # Compute cross product b × c using global temp vector
    cross(crossCD, b, c)
    # Return dot product a · (b × c)
    return dot(a, crossCD)

def vectorDifference(A: "Cartesian", B: "Cartesian") -> float:
    """
    Returns a difference measure between two vectors, a - b
    D = sqrt(1 - dot(a,b)) / sqrt(2)
    D = 1: a and b are perpendicular
    D = 0: a and b are the same
    D = NaN: a and b are opposite (shouldn't happen in IVEA as we're using normalized vectors in the same hemisphere)
    
    D is a measure of the angle between the two vectors. sqrt(2) can be ignored when comparing ratios.
    
    Args:
        A: first vector
        B: second vector
        
    Returns:
        difference measure between A and B
    """
    # Original implementation is unstable for small angles as dot(A, B) approaches 1
    # dot(A, B) = cos(x) as A and B are normalized
    # Using double angle formula for cos(2x) = 1 - 2sin(x)^2, can rewrite as:
    # 1 - cos(x) = 2 * sin(x/2)^2)
    #            = 2 * sin(x/2)^2
    # ⇒ sqrt(1 - cos(x)) = sqrt(2) * sin(x/2) 
    # Angle x/2 can be obtained as the angle between A and the normalized midpoint of A and B
    # ⇒ sin(x/2) = |cross(A, midpointAB)|
    lerp(midpointAB, A, B, 0.5)
    normalize(midpointAB, midpointAB)
    cross(midpointAB, A, midpointAB)
    D = length(midpointAB)

    # Math.sin(x) = x for x < 1e-8
    if D < 1e-8:
        # When A and B are close or equal sin(x/2) ≈ x/2, just take the half-distance between A and B
        subtract(crossCD, A, B)
        half_distance = 0.5 * length(crossCD)
        return half_distance
    return D

def quadrupleProduct(out: Vec3, A: "Cartesian", B: "Cartesian", C: "Cartesian", D: "Cartesian") -> Vec3:
    """
    Computes the quadruple product of four vectors
    
    Args:
        out: output vector
        A, B, C, D: input vectors
        
    Returns:
        out
    """
    cross(crossCD, C, D)
    triple_product_acd = dot(A, crossCD)
    triple_product_bcd = dot(B, crossCD)
    scale(scaledA, A, triple_product_bcd)
    scale(scaledB, B, triple_product_acd)
    return subtract(out, scaledB, scaledA)

def slerp(out: Vec3, A: "Cartesian", B: "Cartesian", t: float) -> "Cartesian":
    """
    Spherical linear interpolation between two vectors
    
    Args:
        out: The target vector to write the result to
        A: The first vector
        B: The second vector
        t: The interpolation parameter (0 to 1)
        
    Returns:
        The interpolated vector (same as out)
    """
    gamma = angle(A, B)
    if gamma < 1e-12:
        return lerp(out, A, B, t)
    
    weight_a = math.sin((1 - t) * gamma) / math.sin(gamma)
    weight_b = math.sin(t * gamma) / math.sin(gamma)
    scale(scaledA, A, weight_a)
    scale(scaledB, B, weight_b)
    add(out, scaledA, scaledB)
    return cast("Cartesian", (out[0], out[1], out[2]))


# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

from .types import YES, NO, Orientation, Quaternary, Flip, Anchor

from .basis import ij_to_kj, kj_to_ij

from .quaternary import quaternary_to_kj, quaternary_to_flips, ij_to_quaternary

from .anchor import compute_q, offset_flips_to_anchor

from .shift_digits import (
    shift_digits, PATTERN, PATTERN_FLIPPED,
    PATTERN_REVERSED, PATTERN_FLIPPED_REVERSED
)

from .hilbert import s_to_anchor, ij_to_s, ij_to_flips, anchor_to_s

from .triple import Triple, triple_parity, triple_in_bounds, triple_to_s, anchor_to_triple, triple_to_anchor



# A5
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) A5 contributors

import json
from pathlib import Path
from a5.lattice import shift_digits, PATTERN, PATTERN_FLIPPED, PATTERN_REVERSED, PATTERN_FLIPPED_REVERSED


def load_fixtures():
    fixture_path = Path(__file__).parent / "fixtures" / "shift-digits.json"
    with open(fixture_path, 'r') as f:
        return json.load(f)


PATTERN_MAP = {
    "PATTERN": PATTERN,
    "PATTERN_FLIPPED": PATTERN_FLIPPED,
    "PATTERN_REVERSED": PATTERN_REVERSED,
    "PATTERN_FLIPPED_REVERSED": PATTERN_FLIPPED_REVERSED,
}


class TestShiftDigits:
    def test_shift_digits_fixtures(self):
        fixtures = load_fixtures()
        for case in fixtures["shiftDigits"]:
            digits = list(case["digitsBefore"])
            flips = list(case["flips"])
            pattern = PATTERN_MAP[case["patternName"]]
            shift_digits(digits, case["i"], flips, case["invertJ"], pattern)
            assert digits == case["digitsAfter"], \
                f'digits {case["digitsBefore"]} -> {digits} != {case["digitsAfter"]}'



# SPDX-License-Identifier: Apache-2.0
import pytest
from a5.core.serialization import (
    serialize,
    deserialize,
    get_resolution,
    MAX_RESOLUTION,
    REMOVAL_MASK,
    FIRST_HILBERT_RESOLUTION,
    WORLD_CELL,
    cell_to_parent,
    cell_to_children,
    get_res0_cells,
)
from a5.core.utils import A5Cell
from a5.core.origin import origins
import json
from pathlib import Path
import copy

# Test data
origin0 = copy.deepcopy(origins[0])
TEST_IDS_PATH = Path(__file__).parent / "test-ids.json"

with open(TEST_IDS_PATH) as f:
    TEST_IDS = json.load(f)

# Resolution masks for testing bit encoding
RESOLUTION_MASKS = [
    # Non-Hilbert resolutions
    "0000001000000000000000000000000000000000000000000000000000000000",  # res0: Dodecahedron faces
    "0000000100000000000000000000000000000000000000000000000000000000",  # res1: Quintants
    # Hilbert resolutions (res2-29)
    "0000000010000000000000000000000000000000000000000000000000000000",
    "0000000000100000000000000000000000000000000000000000000000000000",
    "0000000000001000000000000000000000000000000000000000000000000000",
    "0000000000000010000000000000000000000000000000000000000000000000",
    "0000000000000000100000000000000000000000000000000000000000000000",
    "0000000000000000001000000000000000000000000000000000000000000000",
    "0000000000000000000010000000000000000000000000000000000000000000",
    "0000000000000000000000100000000000000000000000000000000000000000",
    "0000000000000000000000001000000000000000000000000000000000000000",
    "0000000000000000000000000010000000000000000000000000000000000000",
    "0000000000000000000000000000100000000000000000000000000000000000",
    "0000000000000000000000000000001000000000000000000000000000000000",
    "0000000000000000000000000000000010000000000000000000000000000000",
    "0000000000000000000000000000000000100000000000000000000000000000",
    "0000000000000000000000000000000000001000000000000000000000000000",
    "0000000000000000000000000000000000000010000000000000000000000000",
    "0000000000000000000000000000000000000000100000000000000000000000",
    "0000000000000000000000000000000000000000001000000000000000000000",
    "0000000000000000000000000000000000000000000010000000000000000000",
    "0000000000000000000000000000000000000000000000100000000000000000",
    "0000000000000000000000000000000000000000000000001000000000000000",
    "0000000000000000000000000000000000000000000000000010000000000000",
    "0000000000000000000000000000000000000000000000000000100000000000",
    "0000000000000000000000000000000000000000000000000000001000000000",
    "0000000000000000000000000000000000000000000000000000000010000000",
    "0000000000000000000000000000000000000000000000000000000000100000",
    "0000000000000000000000000000000000000000000000000000000000001000",
    "0000000000000000000000000000000000000000000000000000000000000010",
]


# Helper function to compare A5Cell objects
def assert_cells_equal(actual: A5Cell, expected: A5Cell):
    """Compare A5Cell objects properly."""
    assert actual["origin"].id == expected["origin"].id
    assert actual["origin"].axis == expected["origin"].axis
    assert actual["origin"].quat == expected["origin"].quat
    assert actual["origin"].angle == expected["origin"].angle
    assert actual["origin"].orientation == expected["origin"].orientation
    assert actual["origin"].first_quintant == expected["origin"].first_quintant
    assert actual["segment"] == expected["segment"]
    assert actual["S"] == expected["S"]
    assert actual["resolution"] == expected["resolution"]


# =============================================================================
# serialize tests
# =============================================================================

def test_correct_number_of_masks():
    """Test correct number of masks."""
    assert len(RESOLUTION_MASKS) == MAX_RESOLUTION


def test_removal_mask_is_correct():
    """Test removal mask is correct."""
    origin_segment_bits = "0" * 6
    remaining_bits = "1" * 58
    expected = int(f"0b{origin_segment_bits}{remaining_bits}", 2)
    assert REMOVAL_MASK == expected


def test_encodes_resolution_correctly_for_different_values():
    """Test resolution encoding."""
    for i in range(len(RESOLUTION_MASKS)):
        # Origin 0 has first quintant 4, so use segment 4 to obtain start of Hilbert curve
        input_cell = A5Cell(origin=origin0, segment=4, S=0, resolution=i)
        serialized = serialize(input_cell)
        actual_binary = format(serialized, '064b')
        expected_binary = RESOLUTION_MASKS[i]
        assert actual_binary == expected_binary


def test_correctly_extracts_resolution():
    """Test resolution extraction."""
    for i, binary in enumerate(RESOLUTION_MASKS):
        assert len(binary) == 64
        n = int(f"0b{binary}", 2)
        resolution = get_resolution(n)
        assert resolution == i


def test_encodes_origin_segment_and_s_correctly():
    """Test origin, segment and S encoding."""
    # Origin 0 has first quintant 4, so use segment 4 to obtain start of Hilbert curve
    cell = A5Cell(origin=origin0, segment=4, S=0, resolution=MAX_RESOLUTION - 1)
    serialized = serialize(cell)
    assert serialized == 0b10  # 2 in decimal


def test_throws_error_when_s_too_large_for_resolution():
    """Test S too large error."""
    cell = A5Cell(origin=origin0, segment=0, S=16, resolution=3)  # Too large for resolution 3 (max is 15)
    with pytest.raises(ValueError, match="S \\(16\\) is too large for resolution level 3"):
        serialize(cell)


def test_throws_error_when_resolution_exceeds_maximum():
    """Test resolution exceeds maximum error."""
    cell = A5Cell(origin=origin0, segment=0, S=0, resolution=MAX_RESOLUTION + 1)
    with pytest.raises(ValueError, match="Resolution .* is too large"):
        serialize(cell)


# Round trip tests
@pytest.mark.parametrize("id", TEST_IDS)
def test_round_trip_test_ids(id):
    """Test round trip with test IDs."""
    serialized = int(id, 16)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)
    assert reserialized == serialized


@pytest.mark.parametrize("origin_id", range(1, 12))
@pytest.mark.parametrize("binary", RESOLUTION_MASKS[FIRST_HILBERT_RESOLUTION:])
def test_round_trip_resolution_masks_with_origins(origin_id, binary):
    """Test round trip for resolution masks with different origins."""
    origin_segment_id = format(5 * origin_id, '06b')
    serialized = int(f"0b{origin_segment_id}{binary[6:]}", 2)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)
    assert reserialized == serialized


@pytest.mark.parametrize("resolution", range(MAX_RESOLUTION))
def test_serialize_deserialize_round_trip(resolution):
    """Test serialize/deserialize round trip for all resolutions."""
    input_cell = A5Cell(origin=origin0, segment=4, S=0, resolution=resolution)
    
    serialized = serialize(input_cell)
    deserialized = deserialize(serialized)
    
    # At resolution 0, segment is always normalized to 0
    expected_cell = copy.deepcopy(input_cell)
    if resolution == 0:
        expected_cell["segment"] = 0
    
    assert_cells_equal(deserialized, expected_cell)
    assert get_resolution(serialized) == resolution


# =============================================================================
# hierarchy tests
# =============================================================================

@pytest.mark.parametrize("id", TEST_IDS)
def test_cell_to_children_with_same_resolution_returns_original_cell(id):
    cell = int(id, 16)
    current_resolution = get_resolution(cell)
    children = cell_to_children(cell, current_resolution)
    assert len(children) == 1
    assert children[0] == cell

@pytest.mark.parametrize("id", TEST_IDS)
def test_cell_to_parent_with_same_resolution_returns_original_cell(id):
    cell = int(id, 16)
    current_resolution = get_resolution(cell)
    parent = cell_to_parent(cell, current_resolution)
    assert parent == cell

@pytest.mark.parametrize("id", TEST_IDS)
def test_round_trip_between_cell_to_parent_and_cell_to_children(id):
    """Test parent-child round trip."""
    cell = int(id, 16)
    
    children = cell_to_children(cell)
    assert children, "No children returned"
    
    # Test first child
    child = children[0]
    assert cell_to_parent(child) == cell
    
    # Test all children have same parent
    parents = [cell_to_parent(child) for child in children]
    assert all(p == cell for p in parents), "Not all children map to the same parent"


def test_non_hilbert_to_non_hilbert_hierarchy():
    """Test non-Hilbert to non-Hilbert transition."""
    # Test resolution 0 to 1 (both non-Hilbert)
    cell = serialize(A5Cell(origin=origin0, segment=0, S=0, resolution=0))
    children = cell_to_children(cell)
    assert len(children) == 5
    for child in children:
        assert cell_to_parent(child) == cell


def test_non_hilbert_to_hilbert_hierarchy():
    """Test non-Hilbert to Hilbert transition."""
    # Test resolution 1 to 2 (non-Hilbert to Hilbert)
    cell = serialize(A5Cell(origin=origin0, segment=0, S=0, resolution=1))
    children = cell_to_children(cell)
    assert len(children) == 4
    for child in children:
        assert cell_to_parent(child) == cell


def test_hilbert_to_non_hilbert_hierarchy():
    """Test Hilbert to non-Hilbert transition."""
    # Test resolution 2 to 1 (Hilbert to non-Hilbert)
    cell = serialize(A5Cell(origin=origin0, segment=0, S=0, resolution=2))
    parent = cell_to_parent(cell, 1)
    children = cell_to_children(parent)
    assert len(children) == 4
    assert cell in children


def test_low_resolution_hierarchy_chain():
    """Test hierarchy chain."""
    # Test a chain of resolutions from 0 to 4
    resolutions = [0, 1, 2, 3, 4]
    cells = [
        serialize(A5Cell(origin=origin0, segment=0, S=0, resolution=res))
        for res in resolutions
    ]

    # Test parent relationships
    for i in range(1, len(cells)):
        assert cell_to_parent(cells[i]) == cells[i-1]

    # Test children relationships
    for i in range(len(cells) - 1):
        children = cell_to_children(cells[i])
        assert cells[i+1] in children


def test_base_cell_division_counts():
    """Test base cell division."""
    # Start with the base cell (resolution -1)
    base_cell = serialize(A5Cell(origin=origin0, segment=0, S=0, resolution=-1))
    current_cells = [base_cell]
    expected_counts = [12, 60, 240, 960]  # 12, 12*5, 12*5*4, 12*5*4*4

    # Test each resolution level up to 4
    for resolution, expected_count in enumerate(expected_counts):
        # Get all children of current cells
        all_children = []
        for cell in current_cells:
            all_children.extend(cell_to_children(cell))
        
        # Verify the total number of cells matches expected
        assert len(all_children) == expected_count
        
        # Update current cells for next iteration
        current_cells = all_children


# =============================================================================
# getRes0Cells tests
# =============================================================================

def test_get_res0_cells_returns_12_resolution_0_cells():
    """Test getRes0Cells functionality."""
    res0_cells = get_res0_cells()
    assert len(res0_cells) == 12
    
    # Each cell should have resolution 0
    for cell in res0_cells:
        assert get_resolution(cell) == 0



"""
A5
SPDX-License-Identifier: Apache-2.0
Copyright (c) A5 contributors
"""

import math
from typing import cast
from .coordinate_systems import Radians, Polar
from .constants import distance_to_edge, PI_OVER_5, TWO_PI_OVER_5, WARP_FACTOR

def normalize_gamma(gamma: Radians) -> Radians:
    segment = gamma / TWO_PI_OVER_5
    s_center = round(segment)
    s_offset = segment - s_center

    # Azimuthal angle from triangle bisector
    beta = s_offset * TWO_PI_OVER_5
    return cast(Radians, beta)

def _warp_beta(beta: float) -> float:
    shifted_beta = beta * WARP_FACTOR
    return math.tan(shifted_beta)

def _unwarp_beta(beta: float) -> float:
    shifted_beta = math.atan(beta)
    return shifted_beta / WARP_FACTOR

beta_max = PI_OVER_5
WARP_SCALER = _warp_beta(beta_max) / beta_max

def warp_beta(beta: float) -> float:
    return _warp_beta(beta) / WARP_SCALER

def unwarp_beta(beta: float) -> float:
    return _unwarp_beta(beta * WARP_SCALER)

def warp_rho(rho: float, beta: float) -> float:
    beta_ratio = abs(beta) / beta_max
    shifted_rho = rho * (0.95 - 0.05 * beta_ratio)
    return math.tan(shifted_rho)

def unwarp_rho(rho: float, beta: float) -> float:
    beta_ratio = abs(beta) / beta_max
    shifted_rho = math.atan(rho)
    return shifted_rho / (0.95 - 0.05 * beta_ratio)

def warp_polar(polar: Polar) -> Polar:
    rho, gamma = polar
    beta = normalize_gamma(gamma)
    
    beta2 = warp_beta(normalize_gamma(gamma))
    delta_beta = beta2 - beta

    # Distance to edge will change, so shift rho to match
    scale = math.cos(beta) / math.cos(beta2)
    rho_out = scale * rho

    rho_max = distance_to_edge / math.cos(beta2)
    scaler2 = warp_rho(rho_max, beta2) / rho_max
    rho_warped = warp_rho(rho_out, beta2) / scaler2

    return cast(Polar, (rho_warped, gamma + delta_beta))

def unwarp_polar(polar: Polar) -> Polar:
    rho, gamma = polar
    beta2 = normalize_gamma(gamma)
    beta = unwarp_beta(beta2)
    delta_beta = beta2 - beta

    # Reverse the rho warping
    rho_max = distance_to_edge / math.cos(beta2)
    scaler2 = warp_rho(rho_max, beta2) / rho_max
    rho_unwarped = unwarp_rho(rho * scaler2, beta2)
    
    # Reverse the scale adjustment
    scale = math.cos(beta) / math.cos(beta2)
    return cast(Polar, (rho_unwarped / scale, gamma - delta_beta))